"""
TopoConsole class file.


$Id$
"""


# CB: does the status bar need to keep saying 'ok'? Sometimes
# positive feedback is useful, but 'ok' doesn't seem too helpful.


import os
import copy
import sys
import __main__
import webbrowser
import string

from Tkinter import Frame, Button, \
     LEFT, YES, Label, DISABLED, \
     NORMAL, DoubleVar
from tkFileDialog import asksaveasfilename,askopenfilename

import param
from param import normalize_path,resolve_path
import paramtk as tk


import topo
from topo.plotting.plotgroup import plotgroups, FeatureCurvePlotGroup
from topo.misc.keyedlist import KeyedList
from topo.misc.commandline import sim_name_from_filename
import topo.misc.genexamples
import topo.command

import topo.tkgui 

from templateplotgrouppanel import TemplatePlotGroupPanel
from featurecurvepanel import FeatureCurvePanel
from projectionpanel import CFProjectionPanel,ProjectionActivityPanel,ConnectionFieldsPanel,RFProjectionPanel
from testpattern import TestPattern
from editor import ModelEditor


tk.AppWindow.window_icon_path = resolve_path('tkgui/icons/topo.xbm')

SCRIPT_FILETYPES = [('Topographica scripts','*.ty'),
                    ('Python scripts','*.py'),
                    ('All files','*')]
SAVED_FILE_EXTENSION = '.typ'
SAVED_FILETYPES = [('Topographica saved networks',
                    '*'+SAVED_FILE_EXTENSION),
                   ('All files','*')]

turl = "http://topographica.org/"
userman = "User_Manual/index.html"
tuts = "Tutorials/index.html"
refman = "Reference_Manual/index.html"
plotman = "User_Manual/plotting.html"

# for deb on ubuntu; will need to check others
pkgdoc = "/usr/share/doc/topographica/doc/" 

# Documentation locations: locally built and web urls.
user_manual_locations      = ('doc/'+userman,
                              pkgdoc+userman,
                              turl+userman)
tutorials_locations        = ('doc/'+tuts,
                              pkgdoc+tuts,  
                              turl+tuts)
reference_manual_locations = ('doc/'+refman,
                              pkgdoc+refman,
                              turl+refman)
python_doc_locations = ('http://www.python.org/doc/',)
topo_www_locations = (turl,)
plotting_help_locations = ('doc/'+plotman,
                           pkgdoc+plotman,
                           turl+plotman)

# If a particular plotgroup_template needs (or works better with) a
# specific subclass of PlotPanel, the writer of the new subclass
# or the plotgroup_template can declare here that that template
# should use a specific PlotPanel subclass.  For example:
#   plotpanel_classes['Hue Pref Map'] = HuePreferencePanel
plotpanel_classes = {}
# CEBALERT: why are the other plotpanel_classes updates at the end of this file?




def open_plotgroup_panel(class_,plotgroup=None,**kw):

    if class_.valid_context():
        win = topo.guimain.some_area.new_window()
        panel = class_(win,plotgroup=plotgroup,**kw)

        if not panel.dock:
            topo.guimain.some_area.eject(win)
        else:
            topo.guimain.some_area.consume(win)

        panel.refresh_title()

        panel.pack(expand='yes',fill='both')
        win.sizeright()
        
        #frame.sizeright()
        
        #topo.guimain.messageBar.message('state', 'OK')
        return panel
    else:
        topo.guimain.messageBar.response(
            'No suitable objects in this simulation for this operation.')



        

class PlotsMenuEntry(param.Parameterized):
    """
    Stores information about a Plots menu command
    (including the command itself, and the plotgroup template).
    """
    def __init__(self,plotgroup,class_=TemplatePlotGroupPanel,**params):
        """
        Store the template, and set the class that will be created by this menu entry

        If users want to extend the Plot Panel classes, then they
        should add entries to the plotpanel_classes dictionary.
        If no entry is defined there, then the default class is used.

        The class_ is overridden for any special cases listed in this method.
        """
        super(PlotsMenuEntry,self).__init__(**params)

        self.plotgroup = plotgroup

        # Special cases.  These classes are specific to the topo/tkgui
        # directory and therefore this link must be made within the tkgui
        # files.
        if isinstance(self.plotgroup,FeatureCurvePlotGroup):
            class_ = plotpanel_classes.get(self.plotgroup.name,FeatureCurvePanel)

        self.class_ = plotpanel_classes.get(self.plotgroup.name,class_)
        

    def __call__(self,event=None,**kw):
        """
        Instantiate the class_ (used as menu commands' 'command' attribute).

        Keyword args are passed to the class_.
        """
        new_plotgroup = copy.deepcopy(self.plotgroup)

        # CB: hack to share plot_templates with the current
        # plotgroup in plotgroups
        new_plotgroup.plot_templates = topo.plotting.plotgroup.plotgroups[self.plotgroup.name].plot_templates

        return open_plotgroup_panel(self.class_,new_plotgroup,**kw)


# Notebook only available for Tkinter>=8.5
try:
    from paramtk.tilewrapper import Notebook
    class DockManager(Notebook):
        """Manages windows that can be tabs in a notebook, or toplevels."""
        def __init__(self, master=None, cnf={}, **kw):
            Notebook.__init__(self, master, cnf=cnf, **kw)
            self._tab_ids = {}

        def _set_tab_title(self,win,title):
            self.tab(self._tab_ids[win],text=title)

        def _set_toplevel_title(self,win,title):
            prefix = topo.sim.name+": "
            if not title.startswith(prefix):
                title=prefix+title
            self.tk.call("wm","title",win._w,title)

        def add(self, child, cnf={}, **kw):
            self._tab_ids[child]=len(self.tabs())
            Notebook.add(self,child,cnf=cnf,**kw)

    ##     def unhide(self,win):
    ##         if win in self._tab_ids:
    ##             self.tab(self._tab_ids[win],state='normal')

        def new_window(self):
            win = tk.AppWindow(self,status=True)
            #self.consume(win)
            return win

        def consume(self,win):
            if win not in self._tab_ids:
                self.tk.call('wm','forget',win._w)
                win.title = lambda x: self._set_tab_title(win,x)
                self.add(win)

        def eject(self,win):
            if win in self._tab_ids:
                self.forget(self._tab_ids[win])

                # manage my tab ids (HACK)
                del self._tab_ids[win]
                for w in self._tab_ids:
                    self._tab_ids[w]-=1
                    self._tab_ids[w]=max(self._tab_ids[w],0)

                self.tk.call('wm','manage',win._w)
                win.renew()
                win.title = lambda x: self._set_toplevel_title(win,x)
                return win

except ImportError:
    class FakeDockManager(Frame):
        def _set_tab_title(self,*args):
            pass
        def _set_toplevel_title(self,win,title):
            prefix = topo.sim.name+": "
            if not title.startswith(prefix):
                title=prefix+title
            self.tk.call("wm","title",win._w,title)
        def add(self,*args):
            pass
        def new_window(self):
            win = tk.AppWindow(self,status=True)
            return win
        def consume(self,win):
            pass
        def eject(self,win):
            win.renew()
            win.title = lambda x: self._set_toplevel_title(win,x)
            return win

    DockManager = FakeDockManager


# This is really a hack. There doesn't seem to be any easy way to tie
# an exception to the window from which it originated. (I couldn't
# find an example of tkinter software displaying a gui exception on
# the originating window.)
def _tkinter_report_exception(widget):
    exc, val, tb = sys.exc_type, sys.exc_value, sys.exc_traceback
    msg = "(%s) %s"%(exc.__name__,val)
    # If the supplied widget has no master, it's probably the Tk
    # instance. In that case, resort to the 'last-one-set' hack (see
    # CEBALERT "provide a way of allowing other gui components" in
    # topo/param/tk.py).
    if not widget.master:
        widget = tk._last_one_set

    stat = None

    while (widget is not None and widget.master):
        # CEBALERT: should rename all status bars to the same thing
        # (status_bar)
        if hasattr(widget,'status'):
            stat = widget.status
            break
        elif hasattr(widget,'messageBar'):
            stat = widget.messageBar
            break
        widget = widget.master

    if stat is not None:
        stat.error('%s'%msg)
    else:
        topo.guimain.messageBar.error('%s'%msg)
    
    # BK-NOTE: Default is now to display full trace always. Any user
    # errors should be caught as special exception cases 
    
    # BK-ALERT: Want to raise errors vs print, however this currently crashes ipython.
    #raise
    
    param.Parameterized().warning(msg)
    import traceback
    traceback.print_exc()


import Tkinter

class TopoConsole(tk.AppWindow,tk.TkParameterized):
    """
    Main window for the Tk-based GUI.
    """

    def _getmenubar(self):
        return self.master.menubar

    menubar = property(_getmenubar)

    def __getitem__(self,menu_name):
        """Allow dictionary-style access to the menu bar."""
        return self.menubar[menu_name]

    def __init__(self,root,**params):
        tk.AppWindow.__init__(self,root,status=True)
        tk.TkParameterized.__init__(self,root,**params)

        # Instead of displaying tracebacks on the commandline, try to display
        # them on the originating window.
        # CEBALERT: on destroy(), ought to revert this
        Tkinter.Misc._report_exception=_tkinter_report_exception

        self.auto_refresh_panels = []
        self._init_widgets()
        self.title(topo.sim.name) # If -g passed *before* scripts on commandline, this is useless.
                                  # So topo.misc.commandline sets the title as its last action (if -g)



        # catch click on the 'x': offers choice to quit or not
        self.protocol("WM_DELETE_WINDOW",self.quit_topographica)

        
        ##########
        ### Make cascade menus open automatically on linux when the mouse
        ### is over the menu title.
        ### [Tkinter-discuss] Cascade menu issue
        ### http://mail.python.org/pipermail/tkinter-discuss/2006-August/000864.html
        if topo.tkgui.system_platform is 'linux':
            activate_cascade = """\
            if {[%W cget -type] != {menubar} && [%W type active] == {cascade}} {
                %W postcascade active
               }
            """
            self.bind_class("Menu", "<<MenuSelect>>", activate_cascade)
        ##########

        # Install warning and message handling
        from param.parameterized import Parameterized        
        self.__orig_P_warning = Parameterized.warning
        self.__orig_P_message = Parameterized.message
        type.__setattr__(Parameterized,'warning',self.gui_warning)
        type.__setattr__(Parameterized,'message',self.gui_message)

    def gui_warning(self,*args):
        stat = self.__get_status_bar()
        s = string.join(args,' ')
        stat.warn(s)
        self.__orig_P_warning(self,*args)

    def gui_message(self,*args):
        stat = self.__get_status_bar()
        s = string.join(args,' ')
        stat.message(s)
        self.__orig_P_message(self,*args)


    def title(self,t=None):
        newtitle = "Topographica"
        if t: newtitle+=": %s" % t
        tk.AppWindow.title(self,newtitle)
        

    def _init_widgets(self):
        
        ## CEBALERT: now we can have multiple operations at the same time,
        ## status bar could be improved to show all tasks?

        # CEBALERT
        self.messageBar = self.status

        self.some_area = DockManager(self)
        self.some_area.pack(fill="both", expand=1)

        ### Balloon, for pop-up help
        self.balloon = tk.Balloon(self.content)

        ### Top-level (native) menu bar
        #self.menubar = tk.ControllableMenu(self.content)       
        self.configure(menu=self.menubar)

        #self.menu_balloon = Balloon(topo.tkgui.root)

        # no menubar in tile yet
        # http://news.hping.org/comp.lang.tcl.archive/4679.html

        self.__simulation_menu()
        self.__create_plots_menu()
        self.refresh_plots_menu()
        self.__help_menu()

        
        ### Running the simulation
        run_frame = Frame(self.content)
        run_frame.pack(side='top',fill='x',padx=4,pady=8)

        self.run_frame = run_frame
        
        Label(run_frame,text='Run for: ').pack(side=LEFT)
        
        self.run_for_var=DoubleVar()
        self.run_for_var.set(1.0)

        run_for = tk.TaggedSlider(run_frame,
                               variable=self.run_for_var,
                               tag_width=11,
                               slider_length=150,
                               bounds=(0,20000))
        self.balloon.bind(run_for,"Duration to run the simulation, e.g. 0.0500, 1.0, or 20000.")
        run_for.pack(side=LEFT,fill='x',expand=YES)
        run_for.tag.bind("<Return>",self.run_simulation)

        # When return is pressed, the TaggedSlider updates itself...but we also want to run
        # the simulation in this case.
        run_frame.optional_action=self.run_simulation

        go_button = Button(run_frame,text="Go",
                           command=self.run_simulation)
        go_button.pack(side=LEFT)
        
        self.balloon.bind(go_button,"Run the simulation for the specified duration.")

        self.step_button = Button(run_frame,text="Step",command=self.run_step)
        self.balloon.bind(self.step_button,"Run the simulation through the time at which the next events are processed.")
        self.step_button.pack(side=LEFT)
        
        self.sizeright()


    def __simulation_menu(self):
        """Add the simulation menu options to the menubar."""
        simulation_menu = ControllableMenu(self.menubar,tearoff=0)

        self.menubar.add_cascade(label='Simulation',menu=simulation_menu)

        simulation_menu.add_command(label='Run script',command=self.run_script)
        simulation_menu.add_command(label='Save script',command=self.save_script_repr)
        simulation_menu.add_command(label='Load snapshot',command=self.load_snapshot)
        simulation_menu.add_command(label='Save snapshot',command=self.save_snapshot)
        #simulation_menu.add_command(label='Reset',command=self.reset_network)
        simulation_menu.add_command(label='Test Pattern',command=self.open_test_pattern)

        simulation_menu.add_command(label='Model Editor',command=self.open_model_editor)
        simulation_menu.add_command(label='Quit',command=self.quit_topographica)

        

    def open_test_pattern(self):
        return open_plotgroup_panel(TestPattern)


    
    


    def __create_plots_menu(self):
        """
        Add the plot menu to the menubar, with Basic plots on the menu itself and
        others in cascades by category (the plots come from plotgroup_templates).
        """
        plots_menu = ControllableMenu(self.menubar,tearoff=0)
        self.menubar.add_cascade(label='Plots',menu=plots_menu)
        

    # CEBALERT: should split other menus in same way as plots (create/refresh)
    def refresh_plots_menu(self):
        plots_menu = self['Plots']
        plots_menu.delete(0,'end')

        # create menu entries, and get list of categories
        entries=KeyedList() # keep the order of plotgroup_templates (which is also KL)
        categories = []
        for label,plotgroup in plotgroups.items():
            entries[label] = PlotsMenuEntry(plotgroup)
            categories.append(plotgroup.category)
        categories = sorted(set(categories))

        
        # The Basic category items appear on the menu itself.
        assert 'Basic' in categories, "'Basic' is the category for the standard Plots menu entries."
        for label,entry in entries:
            if entry.plotgroup.category=='Basic':
                    plots_menu.add_command(label=label,command=entry.__call__)
                    
        categories.remove('Basic')

        plots_menu.add_separator()
        
        # Add the other categories to the menu as cascades, and the plots of each category to
        # their cascades.
        for category in categories:
            category_menu = ControllableMenu(plots_menu,tearoff=0)
            plots_menu.add_cascade(label=category,menu=category_menu)

            # could probably search more efficiently than this
            for label,entry in sorted(entries):
                if entry.plotgroup.category==category:
                    category_menu.add_command(label=label,command=entry.__call__)

            
        plots_menu.add_separator()

        plots_menu.add_command(label="Help",command=(lambda x=plotting_help_locations: self.open_location(x)))


    def __help_menu(self):
        """Add the help menu options."""

        help_menu = ControllableMenu(self.menubar,tearoff=0,name='help')
        self.menubar.add_cascade(label='Help',menu=help_menu)

        help_menu.add_command(label='About',command=self.new_about_window)
        help_menu.add_command(label="User Manual",
                              command=(lambda x=user_manual_locations: self.open_location(x)))

        help_menu.add_command(label="Tutorials",
                              command=(lambda x=tutorials_locations: self.open_location(x)))

        help_menu.add_command(label="Examples",
                              command=self.run_example_script)
        
        help_menu.add_command(label="Reference Manual",
                              command=(lambda x=reference_manual_locations: self.open_location(x)))
        
        help_menu.add_command(label="Topographica.org",
                              command=(lambda x=topo_www_locations: self.open_location(x)))

        help_menu.add_command(label="Python documentation",
                              command=(lambda x=python_doc_locations: self.open_location(x)))



            
    def quit_topographica(self,check=True,exit_status=0):
        """Quit topographica."""
        if not check or (check and tk.askyesno("Quit Topographica","Really quit?")):
            self.destroy() 

            # matplotlib's tk backend starts its own Tk instances; we
            # need to close these ourselves (at least to avoid error
            # message about 'unusual termination' in Windows).
            try: # not that there should be an error, but just in case...
                import matplotlib._pylab_helpers
                for figman in matplotlib._pylab_helpers.Gcf.get_all_fig_managers():
                    figman.destroy()
            except:
                pass
                
            print "Quit selected; exiting"

            # Workaround for obscure problem on some UNIX systems
            # as of 4/2007, probably including Fedora Core 5.  
            # On these systems, if Topographica is started from a
            # bash prompt and then quit from the Tkinter GUI (as
            # opposed to using Ctrl-D in the terminal), the
            # terminal would suppress echoing of all future user
            # input.  stty sane restores the terminal to sanity,
            # but it is not clear why this is necessary.
            # For more info:
            # http://groups.google.com/group/comp.lang.python/browse_thread/thread/68d0f33c8eb2e02d
            if topo.tkgui.system_platform=="linux" and os.getenv('EMACS')!='t':
                try: os.system("stty sane")
                except: pass
            # CEBALERT: re. above. Shouldn't we be able to store the
            # output of "stty --save" before starting the gui, then
            # ensure that when the gui exits (however badly it
            # happens) run "stty saved_settings"?
                
            # CEBALERT: there was no call to self.master.destroy()
            sys.exit(exit_status)


    def run_script(self):
        """
        Dialog to run a user-selected script

        The script is exec'd in __main__.__dict__ (i.e. as if it were specified on the commandline.)
        """
        script = askopenfilename(initialdir=normalize_path(),filetypes=SCRIPT_FILETYPES)
        if script in ('',(),None): # (representing the various ways no script was selected in the dialog)
            self.messageBar.response('Run canceled')
        else:
            execfile(script,__main__.__dict__)
            self.messageBar.response('Ran ' + script)
            sim_name_from_filename(script)
            self.title(topo.sim.name)

    # CEBALERT: duplicates most of run_script()
    def run_example_script(self):

        
        script = askopenfilename(initialdir=topo.misc.genexamples.find_examples(),
                                 filetypes=SCRIPT_FILETYPES)

        if script in ('',(),None): # (representing the various ways no script was selected in the dialog)
            self.messageBar.response('No example opened')
        else:
            execfile(script,__main__.__dict__)
            self.messageBar.response('Ran ' + script)
            sim_name_from_filename(script)
            self.title(topo.sim.name)

        

    def save_script_repr(self):
        script_name = asksaveasfilename(filetypes=SCRIPT_FILETYPES,
                                        initialdir=normalize_path(),
                                        initialfile=topo.sim.basename()+"_script_repr.ty")
        
        if script_name:
            topo.command.save_script_repr(script_name)
            self.messageBar.response('Script saved to ' + script_name)
            
    
    def load_snapshot(self):
        """
        Dialog to load a user-selected snapshot (see topo.command.load_snapshot() ).
        """
        snapshot_name = askopenfilename(initialdir=normalize_path(),filetypes=SAVED_FILETYPES)

        if snapshot_name in ('',(),None):
            self.messageBar.response('No snapshot loaded.')
        else:
            self.messageBar.dynamicinfo('Loading snapshot (may take some time)...')
            self.update_idletasks()            
            topo.command.load_snapshot(snapshot_name)
            self.messageBar.response('Loaded snapshot ' + snapshot_name)
            self.title(topo.sim.name)

        self.auto_refresh()


    def save_snapshot(self):
        """
        Dialog to save a snapshot (see topo.command.save_snapshot() ).
        
        Adds the file extension .typ if not already present.
        """
        snapshot_name = asksaveasfilename(filetypes=SAVED_FILETYPES,
                                          initialdir=normalize_path(),
                                          initialfile=topo.sim.basename()+".typ")
        
        if snapshot_name in ('',(),None):
            self.messageBar.response('No snapshot saved.')
        else:
            if not snapshot_name.endswith('.typ'):
                snapshot_name = snapshot_name + SAVED_FILE_EXTENSION
                
            self.messageBar.dynamicinfo('Saving snapshot (may take some time)...')
            self.update_idletasks()            
            topo.command.save_snapshot(snapshot_name)
            self.messageBar.response('Snapshot saved to ' + snapshot_name)
    

    def auto_refresh(self):
        """
        Refresh all windows in auto_refresh_panels.
        
        Panels can add and remove themselves to the list; those in the list
        will have their refresh() method called whenever this console's
        autorefresh() is called.
        """
        for win in self.auto_refresh_panels:
            win.refresh()

        self.set_step_button_state()
        self.update_idletasks()

        

    ### CEBERRORALERT: why doesn't updatecommand("display=True") for an
    ### orientation preference map measurement work with the
    ### hierarchical example? I guess this is the reason I thought the
    ### updating never worked properly (or I really did break it
    ### recently - or I'm confused)...
    def refresh_activity_windows(self):
        """
        Update any windows with a plotgroup_key of 'Activity'.

        Used primarily for debugging long scripts that present a lot of activity patterns.
        """
        for win in self.auto_refresh_panels:
            if win.plotgroup.name=='Activity' or win.plotgroup.name=='ProjectionActivity' :
                win.refresh()
                self.update_idletasks()


        

    def open_model_editor(self):
        """Start the Model editor."""
        return ModelEditor(self)




    def new_about_window(self):
        win = tk.AppWindow(self)
        win.withdraw()
        win.title("About Topographica")
        text = Label(win,text=topo.about(display=False),justify=LEFT)
        text.pack(side=LEFT)
        win.deiconify()
        #self.messageBar.message('state', 'OK')
            
    def open_location(self, locations):
        """
        Try to open one of the specified locations in a new window of the default
        browser. See webbrowser module for more information.

        locations should be a tuple.
        """
        # CB: could have been a list. This is only here because if locations is set
        # to a string, it will loop over the characters of the string.
        assert isinstance(locations,tuple),"locations must be a tuple."

        for location in locations:
            try:
                existing_location = resolve_path(location)
                webbrowser.open(existing_location,new=2,autoraise=True)
                self.messageBar.response('Opened local file '+existing_location+' in browser.')
                return ###
            except:
                pass

        for location in locations:
            if location.startswith('http'):
                try:
                    webbrowser.open(location,new=2,autoraise=True)
                    self.messageBar.response('Opened remote location '+location+' in browser.')
                    return ###
                except:
                    pass

        self.messageBar.response("Could not open any of %s in a browser."%locations)


    # CEBALERT: need to take care of removing old messages automatically?
    # (Otherwise callers might always have to pass 'ok'.)
    def status_message(self,m):
        self.messageBar.response(m)

    
    def run_simulation(self,event=None): # event=None allows use as callback
        """
        Run the simulation for the duration specified in the
        'run for' taggedslider.        
        """
        fduration = self.run_for_var.get()
        self.open_progress_window(timer=topo.sim.timer)
        topo.sim.run_and_time(fduration)
        self.auto_refresh()


    # CEBERRORALERT: Step button does strange things at time==0.
    # E.g. for lissom_oo_or, nothing appears to happen. For
    # hierarchical, runs to time==10.
    def run_step(self):

        if not topo.sim.events:
            # JP: step button should be disabled if there are no events,
            # but just in case...
            return

        # JPALERT: This should really use .run_and_time() but it doesn't support
        # run(until=...)
        topo.sim.run(until=topo.sim.events[0].time)
        self.auto_refresh()

    def set_step_button_state(self):
        if topo.sim.events:
            self.step_button.config(state=NORMAL)
        else:
            self.step_button.config(state=DISABLED)


    def __get_status_bar(self,i=2):
        # Hack to find appropriate status bar: Go back through frames
        # until a widget with a status bar is found, and return it.
        try:
            while True:
                f = sys._getframe(i)
                if hasattr(f,'f_locals'):
                    if 'self' in f.f_locals:
                        o = f.f_locals['self']
                        # (temporary hack til ScrolledFrame cleaned up)
                        if o.__class__.__name__!='ScrolledFrame':
                            if hasattr(o,'messageBar'):
                                return o.messageBar
                            elif hasattr(o,'status'):
                                return o.status
                    i+=1
        except:
            pass

        #print "GUI INTERNAL WARNING: failed to determine window on which to display message."
        return self.messageBar

        
    def open_progress_window(self,timer,title=None):
        """
        Provide a convenient link to progress bars.
        """
        stat = self.__get_status_bar()
        return stat.open_progress_window(timer=timer,sim=topo.sim)



# CEBALERT: of course dictionary access is used as an alternative to
# the config method or whatever it's called! So this could cause
# serious confusion to someone trying to set config options using the
# dictionary style access rather than .config()! Either document
# clearly or abandon, and get() and set() to access menu entries by
# name.
class ControllableMenu(tk.Menu):
    """
    A Menu, but where entries are accessible by name (using
    dictionary-style access).

    ** Not truly compatible with Tkinter; work in progress **
    """
    def __getitem__(self,name):
        return self.named_commands[name]



if __name__ != '__main__':
    plotpanel_classes['Connection Fields'] = ConnectionFieldsPanel
    plotpanel_classes['RF Projection'] = RFProjectionPanel
    plotpanel_classes['Projection'] = CFProjectionPanel 
    plotpanel_classes['Projection Activity'] = ProjectionActivityPanel









"""
Unit test for SheetView
$Id$
"""

import unittest

import numpy.oldnumeric as Numeric

from topo.base.sheet import Sheet
from topo.base.boundingregion import BoundingBox
from topo.base.sheetview import *
from topo.plotting.bitmap import PaletteBitmap

# Turn False once development is complete and this module is to be
# included as part of the system unit testing.
DEV = False


### JCALERT! This testfile has to be finished (I would say entirely reviewed)
### It does not really test efficiently has it is now.
### Also, it should be included in tests/__init__.py

class TestSheetView(unittest.TestCase):

    def setUp(self):
        self.s = Sheet()
        self.s.activity = Numeric.array([[1,2],[3,4]])
        self.s2 = Sheet()
        self.s2.activity = Numeric.array([[4,3],[2,1]])

    def test_init(self):

        sv1 = SheetView((self.s.activity,self.s.bounds),
                          src_name=self.s.name)
        # s.sheet_view() returns a SheetView
        self.s.sheet_views['sv1']=sv1
        sv2 = SheetView((self.s.activity,self.s.bounds),
                                 src_name=self.s.name)
        # s.sheet_view() returns a SheetView
        self.s.sheet_views['sv2']=sv2
        
        # Define a type 1 SheetView, with matrix and bounding box.
        sv3 = SheetView((self.s.activity, self.s.bounds))
        sv4 = SheetView((self.s2.activity,self.s2.bounds))
        # Define a type 2 SheetView, 
        sv5 = SheetView((ADD,((sv3,None),(sv4,None))))
        sv6 = SheetView((ADD,[(sv3,None),
                              (sv4,None),
                              (self.s2.activity,self.s2.bounds)]))
        sv7 = SheetView((SUBTRACT,[(sv3,None),
                              (sv4,None),
                              (self.s2.activity,self.s2.bounds)]))

        # Define a type 3 SheetView
        sv8 = SheetView((self.s,'Activity'))
        if DEV:
            sv3.message(sv3.view())
            sv6.message('sv6.debug', sv6._view_list)
            sv6.message(sv6.view())


# CEBALERT: replace with test using FileImage
##     def test_view(self):
##         ImageGenerator.nominal_bounds = BoundingBox(points=((-0.8,-0.8),(0.8,0.8)))
##         ImageGenerator.nominal_density = 100
        
##         input = ImageGenerator(filename='topo/tests/testsheetview.ppm')
##      sv = SheetView((input.activity,input.bounds),
##                           src_name=input.name)
##         input.sheet_views['Activity']=sv
##         sv_tuple = sv.view()
##         map = PaletteBitmap(sv_tuple[0])
##         # map.show()


#     def test_generate_coords(self):
#         sv = UnitViewArray((self.s.activity, self.s.bounds))
#         print sv.generate_coords(1,self.s.bounds)
#         


    def test_sum_maps(self):
        """Stub"""
        self.assertEqual(1+1,2)



# CB: I think this class can be removed now. We have topo/patterns/image.py, which uses a Sheet.

## # CEBALERT: Used in a number of test files. Maybe one day topo/patterns/image.py
## # will be based on a Sheet, in which case this could be removed.
## from numpy.oldnumeric import resize,array
## from topo.base.sheet import Sheet
## from topo.base.simulation import EventProcessor
## from topo.misc.util import NxN
## from param import Parameter
## from pprint import *
## import Image, ImageOps

## class ImageGenerator(Sheet):
##     """

##     parameters:

##       filename = The path to the image file.

##     A sheet that reads a pixel map and uses it to generate an activity
##     matrix.  The image is converted to grayscale and scaled to match
##     the bounds and density of the sheet.

##     NOTE: A bare ImageGenerator only sends a single event, containing
##     its image when it gets the .start() call, to repeatedly generate
##     images, it must have a self-connection.  More elegant, however,
##     would be to convert the ImageGenerator from a sheet to a generator
##     function suitable for use with the GeneratorSheet class (see
##     topo/sheets/generator.py).

##     """
##     filename = param.Parameter(None)
    
##     def __init__(self,**params):

##         super(ImageGenerator,self).__init__(**params)

##         self.verbose("filename = " + self.filename)

##         image = Image.open(self.filename)
##         image = ImageOps.grayscale(image)
##         image = image.resize(self.activity.shape)
##         self.activity = resize(array([x for x in image.getdata()]),
##                                  (image.size[1],image.size[0]))

##      self.verbose("Initialized %s activity from %s" % (NxN(self.activity.shape),self.filename))
##         max_val = float(max(self.activity.ravel()))
##         self.activity = self.activity / max_val


##     def start(self):
##      assert self.simulation
##      self.simulation.enqueue_event_rel(0,self,self,data=self.activity)

##     def input_event(self,src,src_port,dest_port,data):
##         self.send_output(data=self.activity)





suite = unittest.TestSuite()
suite.addTest(unittest.makeSuite(TestSheetView))

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Unit tests for Topographica.

Use the 'run' function to run all the tests.


We use unittest and doctest to create tests. The run() function calls
tests in files in topo/tests/ that:

* have a name beginning with 'test' and ending with '.py', if the file
defines the 'suite' attribute;
* have a name beginning with 'test' and ending with '.txt'.

If Tkinter cannot be imported, files that have a name ending with
'_tk' are not imported (hence any tests that they contain are
skipped).


unittest
========

We use unittest in two different ways. The first is simply
to run a series of tests:


class TestSomething(unittest.TestCase):

    def setUp(self):
        ...

    def test_x(self):
        ...

    def test_y(self):
        ...

    def extra(self):
        ...
        

suite = unittest.TestSuite()
suite.addTest(unittest.makeSuite(TestSomething))


In the example above, setUp will be called, followed by test_x and
test_y (i.e. the methods setUp and test_* are called automatically);
the extra() method will not be called (unless your code calls it).
setUp does not have to exist.


The second way we use unittest is to pass a series of scenarios
through one battery of tests:


class TestSomething(unittest.TestCase):

    def test_x(self):
        ...

    def test_y(self):
        ...


class TestCase1(TestSomething):

    def setUp(self):
        ...


class TestCase2(TestSomething):

    def setUp(self):
        ...


suite = unittest.TestSuite()

cases = [TestScenario1,TestScenario2]
suite.addTests([unittest.makeSuite(case) for case in cases])


In this second example, TestScenario1.setUp will be called, followed
by test_x and test_y. After this, TestScenario2.setUp will be called,
followed again by test_x and test_y. setUp in the two TestScenarios is
therefore used to create some different data or situations to pass
through the tests.



To be run() automatically, unittest files must (a) be named test*.py, and
(b) must define the 'suite' attribute.


Additionally, unittest files should:

(a) contain the following code to allow the file to be run on its own:

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)

    
(b) set suite.requires_display=True if the tests require a DISPLAY.

(c) call topo.tests.start_tkgui() before running tests (e.g. in
    setUP()) if they require the GUI to be running




doctest
=======




$Id$
"""

# CEBALERT: It might be good if tests/ were a directory at the top
# level, with a subdirectory structure mirroring that of topo/. Then
# it is more likely we'd have a separate test file for each module,
# and we could also simply name the files the same as what they are
# testing, which could make it simpler to find the right test file.

# CEBALERT: tests often affect each other. Make sure test authors are
# aware of that, and have some kind of policy.  (Setting class
# attributes, sharing a sim, etc)

# CEBALERT: some of the test modules are missing code to handle running
# (i.e. running as './topographica topo/tests/testsheet.py').



import unittest,doctest,os,re,fnmatch,socket
import param

# Automatically discover all test*.py files in this directory
__all__ = [re.sub('\.py$','',f)
           for f in fnmatch.filter(os.listdir(__path__[0]),'test*.py')]

all_doctest = sorted(fnmatch.filter(os.listdir(__path__[0]),'test*.txt'))

# location in which to create semi-permanent test data
output_path = param.normalize_path.prefix
tests_output_path = os.path.join(output_path,'tests',socket.gethostname())
if not os.path.exists(tests_output_path):
    print "Creating %s"%tests_output_path
    os.makedirs(tests_output_path)



try:
    import Tkinter
except ImportError:
    tk_tests = fnmatch.filter(__all__,'*_tk')
    tk_doctests = fnmatch.filter(all_doctest,'*_tk')
    param.Parameterized().warning('no Tkinter module: skipping %s'%str(tk_tests+tk_doctests))
    for t in tk_tests:
        __all__.remove(t)
    for t in tk_doctests:
        all_doctest.remove(t)
    

try:
    import gmpy
    gmpy_imported=True
except ImportError:
    gmpy_imported=False

if gmpy_imported and gmpy.__file__ is None:
    gmpy_imported=False

if not gmpy_imported:
    import param
    param.Parameterized().warning('no gmpy module: testgmpynumber.txt skipped')
    all_doctest.remove('testgmpynumber.txt')

# CEBALERT: we need to rename these/reorganize the tests
__all__.remove('test_script')
__all__.remove('test_map_measurement')   

try:
    import scikits.audiolab
except ImportError:
    import param
    param.Parameterized().message("no scikits.audiolab: testaudio.py skipped")
    __all__.remove('testaudio')

# CEBALERT: should be using python warnings, and having unittest
# report warnings.
try:
    import matplotlib
except ImportError:
    import param
    param.Parameterized().warning("Matplotlib is not available; skipping Matplotlib tests.")
    __all__.remove('testmatplotlib')
    __all__.remove('testmatplotlib_tk')


__all__.sort()


def all_suite():
    """
    __all__:
    For each test module that defines a 'suite' attribute, add its
    tests.  Only adds tests requiring a display if the DISPLAY
    environment variable is set.

    all_doctest:
    Add each doctest file to the suite.
    """
    suite = unittest.TestSuite()

    for test_name in __all__:
        # import the module
        exec 'import '+test_name

        test_module = locals()[test_name]
        try:        
            print 'Loading suite from module %s ...' % test_name,
            new_test = getattr(test_module,'suite')

            if _check_for_display(new_test):
                print 'ok.'
                suite.addTest(new_test)
            else:
                print 'skipped: No $DISPLAY.'
                
        except AttributeError,err:
            print err

    for filename in all_doctest:
        print 'Loading doctest file', filename
        suite.addTest(doctest.DocFileSuite(filename))
    return suite


# Note that this is set up so that if all the tests are run and
# there's no DISPLAY, tests requiring DISPLAY are skipped - but if a
# test is run individually via run_named() and it requires DISPLAY, an
# error will be raised.
def _check_for_display(suite):
    """
    Return True if no DISPLAY required or DISPLAY is required and it exists,
    otherwise return False.
    """
    if not hasattr(suite,'requires_display'):
        return True
    elif os.getenv('DISPLAY'):
        return True
    else:
        return False    


def run(verbosity=1,test_modules=None):
    """
    Run tests in all test_modules; test_modules defaults to all_suit().

    E.g. to run all tests:
      ./topographica -c 'from topo.tests import run; run()'

    
    verbosity specifies the level of information printed during the
    tests (see unittest.TextTestRunner).


    To run only a subset of the tests, specify a list of test modules or doctest
    file names. For example:

      ./topographica -c 'from topo.tests import run, testimage, testsheet; run(test_modules=[testimage,testsheet,"testDynamicParameter.txt"])'
    """
    import types
         
    if not test_modules:
        run_suite = all_suite()
    else:
        assert isinstance(test_modules,list), 'test_modules argument must be a list of test modules or doctest filenames.'
        
        run_suite = unittest.TestSuite()
        
        for test_module in test_modules:
            if isinstance(test_module,types.ModuleType):
                if _check_for_display(test_module.suite):
                    run_suite.addTest(test_module.suite)
                else:
                    raise Exception("Cannot run test without a valid DISPLAY.")
            elif isinstance(test_module,str):
                if test_module in all_doctest:
                    run_suite.addTest(doctest.DocFileSuite(test_module))
                else:
                    raise ValueError, '"%s" is not an available doctest file.' % test_module
            else:
                raise ValueError, '%s is not a valid test module' % str(test_module)
                                 

    return unittest.TextTestRunner(verbosity=verbosity).run(run_suite)



# CB: if the unit tests were faster, I wouldn't keep needing this...
def run_named(name,verbosity=2):
    """
    Run the named test module.

    Convenience function to make it easy to run a single test module.

    Examples:
      ./topographica -c 'import topo.tests; topo.tests.run_named("testsnapshots.py")'
      ./topographica -c 'import topo.tests; topo.tests.run_named("testDynamicParameter.txt")'
    """
    if name.endswith('.py'):
        module_name = "topo.tests."+name[0:-3]
        import __main__
        exec "import %s"%module_name in __main__.__dict__
        test_module = eval(module_name,__main__.__dict__)
    else:
        test_module = name
    
    run(verbosity,test_modules=[test_module])



"""
Projection and related classes.

$Id$
"""

import numpy
from numpy import array,asarray,ones,sometrue, logical_and, logical_or

import param
from param.parameterized import overridable_property

from topo.misc.keyedlist import KeyedList # CEBALERT: not in base

from sheet import Sheet
from simulation import EPConnection
from functionfamily import TransferFn
from sheetview import ProjectionView

class SheetMask(param.Parameterized):
    """
    An abstract class that defines a mask over a ProjectionSheet object.
    
    This class is typically used for optimization, where mask
    indicates which neurons are active and should be processed
    further. A mask can also be used for lesion experiments, to
    specify which units should be kept inactive.

    See the code for CFProjection and CFResponseFn to see how this
    class can be used to restrict the computation to only those
    neurons that the Mask lists as active.
    """

    # JPALERT: Is there anything about this class that assumes its
    # sheet is a ProjectionSheet?

    def _get_data(self): 
        assert(self._sheet != None)
        return self._data
    def _set_data(self,data): 
        assert(self._sheet != None)
        self._data = data

    data = overridable_property(_get_data,_set_data,doc="""
    Ensure that whenever somebody accesses the data they are not None.""")

    def _get_sheet(self): 
        assert(self._sheet != None)
        return self._sheet
    def _set_sheet(self,sheet): 
        self._sheet = sheet 
        if(self._sheet != None): self.reset()
    
    sheet = overridable_property(_get_sheet,_set_sheet)
    
    def __init__(self,sheet=None,**params):
        super(SheetMask,self).__init__(**params)
        self.sheet = sheet

    def __and__(self,mask):
        return AndMask(self._sheet,submasks=[self,mask])
    def __or__(self,mask):
        return OrMask(self._sheet,submasks=[self,mask])

    # JABALERT: Shouldn't this just keep one matrix around and zero it out,
    # instead of allocating a new one each time?
    def reset(self):
        """Initialize mask to default value (with no neurons masked out)."""
        self.data = ones(self.sheet.shape)

    
    def calculate(self):
        """
        Calculate a new mask based on the activity of the sheet.
        
        For instance, in an algorithm like LISSOM that is based on a
        process of feedforward activation followed by lateral settling,
        the calculation is done at the beginning of each iteration after
        the feedforward activity has been calculated.
        
        Subclasses should override this method to compute some non-default
        mask.
        """
        pass

  
    # JABALERT: Not clear what the user should do with this.
    def update(self):
        """
        Update the current mask based on the current activity and a previous mask.
        
        Should be called only if calculate() has already been called since the last
        reset(); potentially faster to compute than redoing the entire calculate().
        
        Subclasses should override this method to compute some non-default
        mask.
        """
        pass

  
class CompositeSheetMask(SheetMask):
    """
    A SheetMask that computes its value from other SheetMasks.    
    """
    __abstract =  True 

    submasks = param.List(class_=SheetMask)

    def __init__(self,sheet=None,**params):
        super(CompositeSheetMask,self).__init__(sheet,**params)
        assert self.submasks, "A composite mask must have at least one submask."

    def _combine_submasks(self):
        """
        A method that combines the submasks.

        Subclasses should override this method to do their respective
        composite calculations.  The result should be stored in self.data.
        """
        raise NotImplementedError
    
    def _set_sheet(self,sheet):
        for m in self.submasks:
            m.sheet = sheet
        super(CompositeSheetMask,self)._set_sheet(sheet)

    def reset(self):
        for m in self.submasks:
            m.reset()
        self._combine_submasks()
        
    def calculate(self):
        for m in self.submasks:
            m.calculate()
        self._combine_submasks()
            
    def update(self):
        for m in self.submasks:
            m.update()
        self._combine_submasks()



class AndMask(CompositeSheetMask):
    """
    A composite SheetMask that computes its value as the logical AND (i.e. intersection) of its sub-masks.
    """
    def _combine_submasks(self):
        self._data = asarray(reduce(logical_and,(m.data for m in self.submasks)),dtype=int)



class OrMask(CompositeSheetMask):
    """
    A composite SheetMask that computes its value as the logical OR (i.e. union) of its sub-masks.
    """
    def _combine_submasks(self):
        self._data = asarray(reduce(logical_or,(m.data for m in self.submasks)),dtype=int)

    
        
class Projection(EPConnection):
    """
    A projection from a Sheet into a ProjectionSheet.

    Projections are required to support the activate() method, which
    will construct a matrix the same size as the target
    ProjectionSheet, from an input matrix of activity from the source
    Sheet.  Other than that, a Projection may be of any type.
    """
    __abstract=True
    
    strength = param.Number(default=1.0)

    src_port = param.Parameter(default='Activity')
    
    dest_port = param.Parameter(default='Activity')

    output_fns = param.HookList(default=[],class_=TransferFn,doc="""
        Function(s) applied to the Projection activity after it is computed.""")

    plastic = param.Boolean(default=True, doc="""
        Whether or not to update the internal state on each call.
        Allows plasticity to be turned off during analysis, and then re-enabled.""")
    
    activity_group = param.Parameter(default=(0.5,numpy.add), doc="""
       Grouping and precedence specifier for computing activity from
       Projections.  In a ProjectionSheet, all Projections in the
       same activity_group will be summed, and then the results from
       each group will be combined in the order of the activity_group
       using the operator specified by the activity_operator.  For
       instance, if there are two Projections with
       activity_group==(0.2,numpy.add) and two with
       activity_group==(0.6,numpy.divide), activity
       from the first two will be added together, and the result
       divided by the sum of the second two.""")

    # CEBALERT: precedence should probably be defined at some higher level
    # (and see other classes where it's defined, e.g. Sheet)
    precedence = param.Number(default=0.5)
    
       
    def __init__(self,**params):
        super(Projection,self).__init__(**params)
        self.activity = array(self.dest.activity)
        self._plasticity_setting_stack = []

        
    def activate(self,input_activity):
        """
        Compute an activity matrix for output, based on the specified input_activity.
        
        Subclasses must override this method to whatever it means to
        calculate activity in that subclass.
        """
        raise NotImplementedError

    
    def learn(self):
        """
        This function has to be re-implemented by sub-classes, if they wish
        to support learning.
        """
        pass


    def apply_learn_output_fns(self,active_units_mask=True):
        """
        Sub-classes can implement this function if they wish to
        perform an operation after learning has completed, such as
        normalizing weight values across different projections.

        The active_units_mask argument determines whether or not to
        apply the output_fn to non-responding units.
        """
        pass


  
    def override_plasticity_state(self, new_plasticity_state):
        """
        Temporarily override plasticity of medium and long term internal
        state.

        This function should be implemented by all subclasses so that
        it preserves the ability of the Projection to compute
        activity, i.e. to operate over a short time scale, while
        preventing any lasting changes to the state.

        For instance, if new_plasticity_state is False, in a
        Projection with modifiable connection weights, the values of
        those weights should temporarily be made fixed and unchanging
        after this call.  For a Projection with automatic
        normalization, homeostatic plasticity, or other features that
        depend on a history of events (rather than just the current
        item being processed), changes in those properties would be
        disabled temporarily.  Setting the plasticity state to False
        is useful during analysis operations (e.g. map measurement)
        that would otherwise change the state of the underlying
        network.

        Any process that does not have any lasting state, such as
        those affecting only the current activity level, should not
        be affected by this call.

        By default, this call simply calls override_plasticity_state()
        on the Projection's output_fn, and sets the 'plastic'
        parameter to False.
        """
        self._plasticity_setting_stack.append(self.plastic)
        self.plastic=new_plasticity_state

        for of in self.output_fns:
            if hasattr(of,'override_plasticity_state'):
                of.override_plasticity_state(new_plasticity_state)
      

    def restore_plasticity_state(self):
        """
        Restore previous plasticity state of medium and long term
        internal state after a override_plasticity_state call.

        This function should be implemented by all subclasses to
        remove the effect of the most recent override_plasticity_state call,
        e.g. to reenable plasticity of any type that was disabled.
        """
        self.plastic = self._plasticity_setting_stack.pop()

        for of in self.output_fns:
            if hasattr(of,'restore_plasticity_state'):
                of.restore_plasticity_state()


    def get_projection_view(self, timestamp):
        """Returns the activity in a single projection"""
        return ProjectionView((self.activity.copy(),self.dest.bounds),self,timestamp)


    def n_bytes(self):
        """
        Estimate the memory bytes taken by this Projection.

        By default, counts only the activity array, but subclasses
        should implement this method to include also the bytes taken
        by weight arrays and any similar arrays, as a rough lower
        bound from which memory requirements and memory usage patterns
        can be estimated.
        """
        (rows,cols) = self.activity.shape
        return rows*cols
    
    
    def n_conns(self):
        """
        Return the size of this projection, in number of connections.  
        
        Must be implemented by subclasses, if only to declare that no
        connections are stored.
        """         
        raise NotImplementedError



class ProjectionSheet(Sheet):
    """
    A Sheet whose activity is computed using Projections from other sheets.

    A standard ProjectionSheet expects its input to be generated from
    other Sheets. Upon receiving an input event, the ProjectionSheet
    interprets the event data to be (a copy of) an activity matrix
    from another sheet.  The ProjectionSheet provides a copy of this
    matrix to each Projection from that input Sheet, asking each one
    to compute their own activity in response.  The same occurs for
    any other pending input events.

    After all events have been processed for a given time, the
    ProjectionSheet computes its own activity matrix using its
    activate() method, which by default sums all its Projections'
    activity matrices and passes the result through user-specified
    output_fns() before sending it out on the default output port.
    The activate() method can be overridden to sum some of the
    projections, multiply that by the sum of other projections, etc.,
    to model modulatory or other more complicated types of connections.
    """

    dest_ports=['Activity']
    
    src_ports=['Activity']
        
    # CEBALERT: why isn't this a parameter of Sheet?
    # Should be a MaskParameter for safety
    #mask = ClassSelectorParameter(SheetMask,default=SheetMask(),instantiate=True,doc="""
    mask = param.Parameter(default=SheetMask(),instantiate=True,doc="""
        SheetMask object for computing which units need to be computed further.
        The object should be an instance of SheetMask, and will
        compute which neurons will be considered active for the
        purposes of further processing.  The default mask effectively
        disables all masking, but subclasses can use this mask to
        implement optimizations, non-rectangular Sheet shapes,
        lesions, etc.""")

    # CEBALERT: not sure what to call this, and the default should
    # actually be False. True to match existing behavior.  Not sure
    # this parameter is necessary; clean up with NeighborhoodMask.
    # (Plus same comment as mask: why not parameter of sheet?)
    allow_skip_non_responding_units = param.Boolean(default=True,doc="""
        If true, then units that are inactive after the response
        function has been called can be skipped in subsequent
        processing. Whether or not the units will actually be skipped
        depends on the implementation of learning and learning output
        functions.""")

    
    def __init__(self,**params):
        super(ProjectionSheet,self).__init__(**params)
        self.new_input = False
        self.mask.sheet = self
        self.old_a = self.activity.copy()*0.0

    def _dest_connect(self, conn):
        """
        See EventProcessor's _dest_connect(); raises an error if conn is not
        a Projection.  Subclasses of ProjectionSheet that know how to handle
        other types of Connections should override this method.
        """
        if isinstance(conn, Projection):
            super(ProjectionSheet,self)._dest_connect(conn)
        else:
            raise TypeError('ProjectionSheets only accept Projections, not other types of connection.')


    def input_event(self,conn,data):
        """
        Accept input from some sheet.  Call .present_input() to
        compute the stimulation from that sheet.
        """
        self.verbose("Received input from %s on dest_port %s via connection %s." % (conn.src.name,str(conn.dest_port),conn.name))
        self.present_input(data,conn)
        self.new_input = True


    def _port_match(self,key,portlist):
        """
        Returns True if the given key matches any port on the given list.

        A port is considered a match if the port is == to the key,
        or if the port is a tuple whose first element is == to the key,
        or if both the key and the port are tuples whose first elements are ==.

        This approach allows connections to be grouped using tuples.
        """
        port=portlist[0]
        return [port for port in portlist
                if (port == key or
                    (isinstance(key,tuple)  and key[0] == port) or
                    (isinstance(port,tuple) and port[0] == key) or
                    (isinstance(key,tuple)  and isinstance(port,tuple) and port[0] == key[0]))]


    def _grouped_in_projections(self,ptype):
        """
        Return a dictionary of lists of incoming Projections, grouped by type.

        Each projection of type <ptype> is grouped according to the
        name of the port, into a single list within the dictionary.

        The entry None will contain those that are not of type
        <ptype>, while the other entries will contain a list of
        Projections, each of which has type ptype.
        
        Example: to obtain the lists of projections that should be
        jointly normalised together, call
        __grouped_in_projection('JointNormalize').
        """
        in_proj = KeyedList()
        in_proj[None]=[] # Independent (ungrouped) connections
        
        for c in self.in_connections:
            d = c.dest_port
            if not isinstance(c,Projection):
                self.debug("Skipping non-Projection "+c.name)
            elif isinstance(d,tuple) and len(d)>2 and d[1]==ptype:
                if in_proj.get(d[2]):
                    in_proj[d[2]].append(c)
                else:
                    in_proj[d[2]]=[c]
            #elif isinstance(d,tuple):
            #    raise ValueError("Unable to determine appropriate action for dest_port: %s (connection %s)." % (d,c.name))
            else:
                in_proj[None].append(c)
                    
        return in_proj


    def activate(self):
        """
        Collect activity from each projection, combine it to calculate
        the activity for this sheet, and send the result out. 

        Subclasses may override this method to whatever it means to
        calculate activity in that subclass.
        """
        
        self.activity *= 0.0
        tmp_dict={}
        
        for proj in self.in_connections:
            if (proj.activity_group != None) | (proj.dest_port[0] != 'Activity'):
                if not tmp_dict.has_key(proj.activity_group[0]): 
                    tmp_dict[proj.activity_group[0]]=[]
                tmp_dict[proj.activity_group[0]].append(proj)
        
        keys = tmp_dict.keys()
        keys.sort()
        for priority in keys:
            tmp_activity = self.activity.copy() * 0.0
            
            for proj in tmp_dict[priority]:
                tmp_activity += proj.activity
            self.activity=tmp_dict[priority][0].activity_group[1](self.activity,tmp_activity)
        
        if self.apply_output_fns:
            for of in self.output_fns:
                of(self.activity)
    
        self.send_output(src_port='Activity',data=self.activity)
    

    def process_current_time(self):
        """
        Called by the simulation after all the events are processed for the 
        current time but before time advances.  Allows the event processor
        to send any events that must be sent before time advances to drive
        the simulation. 
        """
        if self.new_input:
            self.activate()
            self.new_input = False
            if self.plastic:
                self.learn()


    def learn(self):
        """
        By default, call the learn() and apply_learn_output_fns()
        methods on every Projection to this Sheet.
        
        Any other type of learning can be implemented by overriding this method.
        Called from self.process_current_time() _after_ activity has
        been propagated.
        """
        for proj in self.in_connections:
            if not isinstance(proj,Projection):
                self.debug("Skipping non-Projection "+proj.name)
            else:
                proj.learn()
                proj.apply_learn_output_fns()


    def present_input(self,input_activity,conn):
        """
        Provide the given input_activity to each in_projection that has a dest_port
        equal to the specified port, asking each one to compute its activity.
        
        The sheet's own activity is not calculated until activate()
        is called.
        """
        conn.activate(input_activity)


    def projections(self,name=None):
        """
        
        Return either a named input p, or a dictionary
        {projection_name, projection} of all the in_connections for
        this ProjectionSheet.

        A minor convenience function for finding projections by name;
        the sheet's list of in_connections usually provides simpler
        access to the Projections.
        """
        if not name:
            return dict([(p.name,p) for p in self.in_connections])
        else:
            for c in self.in_connections:
                if c.name == name:
                    return c
            raise KeyError(name)


    def override_plasticity_state(self, new_plasticity_state):
        """
        Temporarily override plasticity state of medium and long term
        internal state.

        This function should be implemented by all subclasses so that
        when new_plasticity_state=False, it preserves the ability of
        the ProjectionSheet to compute activity, i.e. to operate over
        a short time scale, while preventing any lasting changes to
        the state.

        Any process that does not have any lasting state, such as
        those affecting only the current activity level, should not
        be affected by this call.

        By default, calls override_plasticity_state() on the
        ProjectionSheet's output_fns and all of its incoming
        Projections, and also enables the 'plastic' parameter for this
        ProjectionSheet.  The old value of the plastic parameter is
        saved to an internal stack to be restored by
        restore_plasticity_state().
        """
        super(ProjectionSheet,self).override_plasticity_state(new_plasticity_state)

        for of in self.output_fns:
            if hasattr(of,'override_plasticity_state'):
                of.override_plasticity_state(new_plasticity_state)
                
        for proj in self.in_connections:
            # Could instead check for a override_plasticity_state method
            if isinstance(proj,Projection):
                proj.override_plasticity_state(new_plasticity_state)


    def restore_plasticity_state(self):
        super(ProjectionSheet,self).restore_plasticity_state()

        for of in self.output_fns:
            if hasattr(of,'restore_plasticity_state'):
                of.restore_plasticity_state()

        for proj in self.in_connections:
            if isinstance(proj,Projection):
                proj.restore_plasticity_state()
        
    
    def n_bytes(self):
        """
        Estimate the memory bytes taken by this Sheet and its Projections.

        Typically, this number will include the activity array and any
        similar arrays, plus memory taken by all incoming Projections.
        It will not usually include memory taken by the Python
        dictionary or various "housekeeping" attributes, which usually
        contribute only a small amount to the memory requirements.
        Thus this value should be considered only a rough lower bound
        from which memory requirements and memory usage patterns can
        be estimated.

        Subclasses should reimplement this method if they store a
        significant amount of data other than in the activity array
        and the projections.
        """
        return self.activity.nbytes + \
               sum([p.n_bytes() for p in self.in_connections
                    if isinstance(p,Projection)])
    
    
    def n_conns(self):
        """
        Count the total size of all incoming projections, in number of connections.  
        """
        return sum([p.n_conns() for p in self.in_connections
                    if isinstance(p,Projection)])


# CEBALERT: untested
class NeighborhoodMask(SheetMask):
    """
    A SheetMask where the mask includes a neighborhood around active neurons.

    Given a radius and a threshold, considers a neuron active if at
    least one neuron in the radius is over the threshold.
    """

    threshold = param.Number(default=0.00001,bounds=(0,None),doc="""
       Threshold for considering a neuron active.
       This value should be small to avoid discarding significantly active
       neurons.""")

    radius = param.Number(default=0.05,bounds=(0,None),doc="""
       Radius in Sheet coordinates around active neurons to consider
       neighbors active as well.  Using a larger radius ensures that
       the calculation will be unaffected by the mask, but it will
       reduce any computational benefit from the mask.""")

    
    def __init__(self,sheet,**params):
        super(NeighborhoodMask,self).__init__(sheet,**params)


    def calculate(self):
        rows,cols = self.data.shape
        
        # JAHACKALERT: Not sure whether this is OK. Another way to do
        # this would be to ask for the sheet coordinates of each unit
        # inside the loop.

        # transforms the sheet bounds with bounds2slice() and then
        # uses this to cut out the activity window
        ignore1,matradius = self.sheet.sheet2matrixidx(self.radius,0)
        ignore2,x = self.sheet.sheet2matrixidx(0,0)
        matradius = abs(matradius-x)
        for r in xrange(rows):
            for c in xrange(cols):
                rr = max(0,r-matradius)
                cc = max(0,c-matradius)
                neighbourhood = self.sheet.activity[rr:r+matradius+1,cc:c+matradius+1].ravel()
                self.data[r][c] = sometrue(neighbourhood>self.threshold)



"""
Tests for the ParametersFrame classes.

$Id$
"""


import __main__
import unittest
import Tkinter

import param
import paramtk as tk

from topo.base.patterngenerator import PatternGenerator



# CEBALERT: will be replaced with call to param.tk.initialize() when
# param.tk doesn't depend on anything from tkgui.__init__ (plus this
# test will move to the param package...)
import os
if os.getenv("DISPLAY"):
    import topo.tkgui
    topo.tkgui.start()


class TestPO(param.Parameterized):
    boo = param.Boolean(default=True)
    osp = param.ObjectSelector()
    csp = param.ClassSelector(class_=PatternGenerator)
    const = param.Parameter(1.0,constant=True)
    pa = param.Parameter(default="test")
    nu = param.Number(default=1.0,bounds=(-1,1))
    st = param.String("string1")



class TestParametersFrameWithApply(unittest.TestCase):


    def setUp(self):

        self.some_pos = [param.Parameterized(),param.Parameterized(),
                         param.Parameterized()]
        
        self.testpo1 = TestPO()
        self.testpo1.params()['osp'].objects = self.some_pos

        self.testpo2 = TestPO()
        self.testpo1.params()['osp'].objects = self.some_pos
        
        self.toplevel = Tkinter.Toplevel()
        self.f = tk.ParametersFrameWithApply(self.toplevel,self.testpo1)



    def test_set_PO(self):
        self.assertEqual(self.f._extraPO,self.testpo1)
        self.f.set_PO(self.testpo2)
        self.assertEqual(self.f._extraPO,self.testpo2)
        


    def test_apply_button_1(self):
        """
        Check:
          when pressing apply after no changes, objects should remain
          unchanged, and should still be displayed the same as before.
        """
        ## check that update doesn't affect the variable values
        orig_values = {}
        for param_name,tkvar in self.f._tkvars.items():
            orig_values[param_name] = tkvar._original_get()

        self.f.Apply()

        for param_name,val in orig_values.items():
            self.assertEqual(self.f._tkvars[param_name].get(),val)

        ## and check that *displayed* values don't change 
        orig_values = {}
        for param_name,representation in self.f.representations.items():
            widget = representation['widget']
            
            # button-type widgets don't have a value
            widget_has_value = not 'command' in widget.config()
                            
            if widget_has_value:
                orig_values[param_name] = widget.get()
        
        self.f.Apply()

        for param_name,val in orig_values.items():
            widget = self.f.representations[param_name]['widget']
            self.assertEqual(widget.get(),val)


    def test_apply_button_2(self):
        """
        Check:
          When code to instantiate an object is typed into the entry
          representing a Parameter, if that object exists in main it
          should be instantiated. Further, the display should NOT then
          include quote marks.

          (Only time should get quotation marks is when editing a
          Parameter and trying to create an instance of a class that
          hasn't been imported: then string is assumed, and quotes are
          added - since it's not a StringParameter.)

          Finally, check that when the string remains the same in a box
          that a new object is not instantiated each time Apply is pressed.
        """
        exec "from topo.tests.testparametersframe_tk import TestPO" in __main__.__dict__
        w = self.f.representations['pa']['widget']
        w.delete(0,"end")
        w.insert(0,"TestPO()")
        self.f.Apply()
        content = w.get()
        self.assertEqual(type(self.f.pa),TestPO) # get the object?
        self.assertEqual(content[0:7],"TestPO(") # object displayed right?

        # Now check that pressing apply over and over does not
        # create a new object each time when the same string remains
        # in the widget 
        testpo_id = id(self.f.pa)
        self.f.Apply()
        self.assertEqual(id(self.f.pa),testpo_id)
        # ...but that we do get a new object when the string really changes
        w.delete(0,"end")
        w.insert(0,"TestPO(name='fred')")
        self.f.Apply()
        self.assertNotEqual(id(self.f.pa),testpo_id)


    # (CB: this isn't the place for this test)
    def test_link_to_topo_sim(self):
        """Indicate there's a bug. Have yet to investigate where it actually comes from."""
        import topo
        from utils import new_simulation
        new_simulation()
        e = tk.edit_parameters(topo.sim['S'])
        e.gui_set_param('precedence',7)
        e.Apply()
        self.assertEqual(topo.sim['S'].precedence,7)
        # problem with Simulation no longer being singleton?
        
        

# CEBALERT: need test for defaults button. Check that works ok
# for non-instantiated params etc

            
# test apply disabling etc
    
        



###########################################################

cases = [TestParametersFrameWithApply]

suite = unittest.TestSuite()
suite.addTests([unittest.makeSuite(case) for case in cases])
suite.requires_display = True

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)






###################################################################################
# SNAPSHOT STUFF
###################################################################################

import inspect
import __main__

from param.parameterized import Parameterized, Parameter

# CEBALERT: Can't this stuff move to the ParameterizedMetaclass?
class PicklableClassAttributes(object):
    """
    Supports pickling of Parameterized class attributes for a given module.

    When requested to be pickled, stores a module's PO classes' attributes,
    and any given startup_commands. On unpickling, executes the startup
    commands and sets the class attributes.
    """

    # classes that aren't parameterized any more
    do_not_restore = []


    # Support for changing parameter names
    # CEBALERT: doesn't support things like changing output_fn to output_fns,
    # where we also need to do output_fn=x -> output_fns=[x].
    # Should implement fuller support in legacy, and remove this from here.
    param_name_changes = {}
    # e.g. you change topo.pattern.Gaussian.aspect_ratio to aspect_ration
    # _param_name_changes['topo.pattern.Gaussian']={'aspect_ratio':'aspect_ration'}
    #
    # (not yet finished - do we need to add information about version numbers?)

    # CEBALERT: same comments as above about doing this more cleanly
    param_moves = {}

    # pylint: disable-msg=R0903
    
    # CB: might have mixed up module and package in the docs.
    def __init__(self,module,exclusions=(),startup_commands=()):
        """
        module: a module object, such as topo
        
        Any submodules listed by name in exclusions will not have their
        classes' attributes saved.
        """
        self.module=module
        self.exclude=exclusions
        self.startup_commands=startup_commands
        
    def __getstate__(self):
        """
        Return a dictionary of self.module's PO classes' attributes, plus
        self.startup_commands.
        """
        class_attributes = {}
        self.get_PO_class_attributes(self.module,class_attributes,[],exclude=self.exclude)

        # CB: we don't want to pickle anything about this object except what
        # we want to have executed on unpickling (this object's not going to be hanging around).
        return {'class_attributes':class_attributes,
                'startup_commands':self.startup_commands}

    def __setstate__(self,state):
        """
        Execute the startup commands and set class attributes.
        """
        self.startup_commands = state['startup_commands']
        
        for cmd in self.startup_commands:
            exec cmd in __main__.__dict__

        to_restore = {}

        ########## pre-processing (renames, moves, etc)
        for class_path,state in state['class_attributes'].items():
            # from e.g. "topo.base.parameter.Parameter", we want "topo.base.parameter"

            if class_path in self.do_not_restore:
                #print "Did not restore:",class_path
                break

            for p_name,p_obj in state.items():
                if p_name in self.param_moves.get(class_path,{}):
                    assert p_name not in self.param_name_changes.get(class_path,{})

                    new_class_path,new_p_name = self.param_moves[class_path][p_name]

                    if new_class_path not in to_restore:
                        to_restore[new_class_path] = {}

                    Parameterized().message("%s.%s has been moved to %s.%s"%(class_path,p_name,new_class_path,new_p_name))
                    assert new_p_name not in to_restore[new_class_path]
                    to_restore[new_class_path][new_p_name]=p_obj

                    
                elif p_name in self.param_name_changes.get(class_path,{}):                    
                    new_p_name = self.param_name_changes[class_path][p_name]

                    if class_path not in to_restore:
                        to_restore[class_path] = {}

                    Parameterized().message("%s's %s parameter has been renamed to %s."%(class_path,p_name,new_p_name))
                    to_restore[class_path][new_p_name] = p_obj

                else:
                    if class_path not in to_restore:
                        to_restore[class_path] = {}
                    to_restore[class_path][p_name]= p_obj
                        

        ########## restoring
        for class_path in to_restore:
            module_path = class_path[0:class_path.rindex('.')]
            class_name = class_path[class_path.rindex('.')+1::]

            try:
                module = __import__(module_path,fromlist=[module_path])
            except:
                Parameterized().warning("Could not find module '%s' to restore parameter values of '%s' (module might have been moved or renamed; if you are using this module, please file a support request via topographica.org"%(module_path,class_path))
                break

            try:
                class_=getattr(module,class_name)
            except:
                Parameterized().warning("Could not find class '%s' to restore its parameter values (class might have been removed or renamed; if you are using this class, please file a support request via topographica.org)."%class_path)
                break

            for p_name,p_obj in to_restore[class_path].items():

                if p_name not in class_.params():
                    # CEBALERT: GlobalParams's source code never has
                    # parameters. If we move Parameter saving and
                    # restoring to Parameterized, could allow
                    # individual classes to customize Parameter
                    # restoration.
                    if class_.__name__!='GlobalParams':
                        Parameterized(name='load_snapshot').warning("%s.%s found in snapshot, but '%s' is no longer defined as a Parameter by the current version of %s. If you are using this class, please file a support request via topographica.org." % (class_.__name__, p_name,p_name,class_.__name__))
                else:
                    setattr(class_,p_name,p_obj)
                                        

    # CB: I guess this could be simplified
    def get_PO_class_attributes(self,module,class_attributes,processed_modules,exclude=()):
        """
        Recursively search module and get attributes of Parameterized classes within it.

        class_attributes is a dictionary {module.path.and.Classname: state}, where state
        is the dictionary {attribute: value}.

        Something is considered a module for our purposes if inspect says it's a module,
        and it defines __all__. We only search through modules listed in __all__.

        Keeps a list of processed modules to avoid looking at the same one
        more than once (since e.g. __main__ contains __main__ contains
        __main__...)

        Modules can be specifically excluded if listed in exclude.
        """
        dict_ = module.__dict__
        for (k,v) in dict_.items():
            if '__all__' in dict_ and inspect.ismodule(v) and k not in exclude:
                if k in dict_['__all__'] and v not in processed_modules:
                    self.get_PO_class_attributes(v,class_attributes,processed_modules,exclude)
                    processed_modules.append(v)

            else:
                if isinstance(v,type) and issubclass(v,Parameterized):

                    # Note: we take the class name as v.__name__, not
                    # k, because k might be just a label for the true
                    # class. For example, if someone imports a class
                    # using 'as', the name in the local namespace
                    # could be different from the name when the class
                    # was defined.  It is correct to set the
                    # attributes on the true class.
                    full_class_path = v.__module__+'.'+v.__name__
                    class_attributes[full_class_path] = {}
                    # Parameterized classes always have parameters in
                    # __dict__, never in __slots__
                    for (name,obj) in v.__dict__.items():
                        if isinstance(obj,Parameter) and obj.pickle_default_value:
                            class_attributes[full_class_path][name] = obj




"""
Code used to support old snapshots (those created from 0.9.7/r11275
onwards).

$Id$
"""
__version__='$Revision: 8021 $'

import sys
import decimal # CEBALERT: when did decimal appear? too late to use?
    
import param

from snapshots import PicklableClassAttributes

# CEBALERT: remove the extraneous "import param"s

# CEB: Add note that snapshot can be re-saved, making updates
# permanent. All functions in here should be written to support that.

# CEB: code to support older snapshots is available in earlier
# versions of this file (e.g. r11323). I consider that version of the
# file to be a "proof of concept"; techniques from it will be used to
# support changes made from 0.9.7 onwards.

# CEBNOTE: could also support running old scripts without modifying
# them.
#
# E.g. we changed output_fn to output_fns for several classes. To
# support that in a snapshot requires intercepting the saved data and
# detecting 'output_fn=x' and replacing it with 'output_fns=[x]',
# etc. To support it in a script requires installing something in the
# class that will take .output_fn=x and actually do
# .output_fns=[x]. Here's one way we could support CFProjection's
# weights_shape changing to cf_shape:
#
#    import topo.base.cf
#    cfp = topo.base.cf.CFProjection
#    type.__setattr__(cfp,'weights_shape',cfp.__dict__['cf_shape'])


# CEBALERT: should probably restructure this file so that as little as
# possible happens on import. Should be easy to do.

releases = {"0.9.7": 11275}


def _get_version(snapshot_release,snapshot_version):

    found_version = False
    
    if snapshot_version is not None:
        snapshot_version = snapshot_version.split(":")[0]
        snapshot_version = snapshot_version.split("M")[0]
        
        if len(snapshot_version)>0:
            try:
                snapshot_version = int(snapshot_version)
                found_version = True
            except ValueError:
                pass
        
    if not found_version:
        snapshot_version = releases[snapshot_release]
        param.Parameterized().debug("No version could be detected for this snapshot; assuming version of release %s (i.e. %s)."%(snapshot_release,snapshot_version))

    return snapshot_version


        
class SnapshotSupport(object):

    @staticmethod
    def install(snapshot_release,snapshot_version=None):

        # CEB: I think there's no simple way to tell what "version" of
        # Topographica a snapshot comes from. When you're running
        # Topographica from svn, you can try topo.version, but you'll
        # get things like 11499:11503 or 11499M. If you use git,
        # you'll see "exported". Therefore, we can't always have
        # fine-grained control over what's loaded. We can at least use
        # the release number for coarse-grained control, though.

        snapshot_version = _get_version(snapshot_release,snapshot_version)

        param.Parameterized().debug("Snapshot is from release %s (r%s)"%(snapshot_release,snapshot_version))

        SnapshotSupport.apply_external_patches()
        SnapshotSupport.apply_support(snapshot_version)


    @staticmethod
    def apply_external_patches():
        global external_patches

        # not ordered
        for message in external_patches.keys():
            #param.Parameterized().message(message)
            external_patches[message]()


    @staticmethod
    def apply_support(version_to_support):
        global support
        
        # apply oldest to newest
        for version in sorted(support.keys())[::-1]:
            if version_to_support < version:
                param.Parameterized().message("Applying legacy support for change r%s"%version)
                support[version]()



######################################################################
######################################################################

# Supporting code

def _version_greater_or_equal(module,than):
    # CB: I forgot Python: how do I do "if module version >= than"?!
    major,minor = module.__version__.split(".")[0:2]
    version = decimal.Decimal(major+"."+minor)
    return version>=than

def _setstate(inst,state):
    for k,v in state.items():
        setattr(inst,k,v)

def preprocess_state(class_,state_mod_fn): 
    """
    Allow processing of state with state_mod_fn before
    class_.__setstate__(instance,state) is called.

    state_mod_fn must accept two arguments: instance and state.
    """
    if not hasattr(class_,'__setstate__'):
        # e.g. class_ used to be Parameterized, but now isn't
        class_.__setstate__ = _setstate

    old_setstate = class_.__setstate__
    def new_setstate(instance,state):
        state_mod_fn(instance,state) 
        old_setstate(instance,state)
    class_.__setstate__ = new_setstate

# CEBNOTE: eventually, might have to support multiple redirections for
# one module (see version in r11323).
def module_redirect(name,parent,actual_module):
    """
    For use when module parent.name is now actual_module.
    """
    sys.modules[parent.__name__+'.'+name]=actual_module
    setattr(sys.modules[parent.__name__],name,actual_module)

# CEBALERT: adapted from topo.misc.util's version; need to merge and
# clean up. Also, needs a better name: used only for allow_import().
class _ModuleFaker(object):
    def __init__(self,module):
        self.module = module
        
    def load_module(self,name):
        if name not in sys.modules:
            self.module.__file__ = self.path
            sys.modules[name] = self.module
            if '.' in name:
                parent_name, child_name = name.rsplit('.', 1)
                setattr(sys.modules[parent_name], child_name, self.module)
        return sys.modules[name]
    
class _ModuleImporter(object):
    def __init__(self,module,fullname):
        self.fullname = fullname
        self.module = module
    
    def find_module(self, fullname, path=None):
        if fullname == self.fullname:
            param.Parameterized().message("%s imported as %s"%(self.module.__name__,self.fullname))
            faker = _ModuleFaker(self.module)
            faker.path = path
            return faker
        return None

def allow_import(module,location):
    """Allow module to be imported from location."""
    sys.meta_path.append(_ModuleImporter(module,location))


######################################################################
######################################################################

# Functions to update old snapshots

# support[v]=fn : for snapshots saved before v, call fn

# only ONE entry per svn revision
support = {}

def do_not_restore_paths():
    # For snapshots saved before 11323
    # Avoid restoring search_paths,prefix for resolve_path,normalize_path
    # (For snapshots before r11323, these were included.)    
    PicklableClassAttributes.do_not_restore+=[
        'param.normalize_path',
        'param.resolve_path']

support[11323] = do_not_restore_paths

    
def param_add_pickle_default_value():
    # For snapshots saved before 11321
    # pickle_default_value attribute added to Parameter in r11321

    from topo import param
    def _param_add_pickle_default_value(instance,state):
        if 'pickle_default_value' not in state:
            state['pickle_default_value']=True
    preprocess_state(param.Parameter,_param_add_pickle_default_value)

support[11321] = param_add_pickle_default_value


######################################################################
# CEB: deliberately no support for audio-related changes, since
# audio-related code is changing fast and isn't in general
# use. Support could be added if necessary.

def pattern_basic_rectangular_removed():
    # 11558 pattern.basic.rectangular() was removed
    # CB: I'm assuming nobody cares about this, but if
    # they do, replace the lambda with 11557's rectangular()
    def rectangular(*args,**kw): raise NotImplementedError
    import topo.pattern as B
    B.rectangular = rectangular

support[11558] = pattern_basic_rectangular_removed

# CEBALERT: should be renamed (it's not only about pattern.basic). And
# did all these happen in one commit?  I added topo.ep recently; did
# that happen in the same commit as the others or not?
def pattern_basic_removed():
   import topo.pattern
   import topo.command
   import topo.coordmapper
   import topo.ep
   import topo.learningfn
   import topo.numbergen
   import topo.pattern
   import topo.projection
   import topo.responsefn
   import topo.sheet
   import topo.transferfn
   module_redirect('basic',topo.command,topo.command)
   module_redirect('basic',topo.coordmapper,topo.coordmapper)
   module_redirect('basic',topo.ep,topo.ep)
   module_redirect('basic',topo.learningfn,topo.learningfn)
   module_redirect('basic',topo.numbergen,topo.numbergen)
   module_redirect('basic',topo.pattern,topo.pattern)
   module_redirect('basic',topo.projection,topo.projection)
   module_redirect('basic',topo.responsefn,topo.responsefn)
   module_redirect('basic',topo.sheet,topo.sheet)
   module_redirect('basic',topo.transferfn,topo.transferfn)

support[11871] = pattern_basic_removed


def param_external_removed():
    # CB: From param/external.py, only odict should be relevant to snapshots.
    import topo.misc.odict
    allow_import(topo.misc.odict,'param.external')
            
support[12024] = param_external_removed


# CEBALERT: slot removal/addition support could be extracted to its
# own function, then there'd be less duplication and it'd be easier to
# add new slot support.
def Number_and_BoundingRegion_add_set_hook():
    import param
    import topo.base.boundingregion as R
    def _add_set_hook(instance,state):
        if 'set_hook' not in state and hasattr(instance.__class__,'set_hook'):
            # have to add to state or else slot won't exist on instance, but will
            # exist on class (consequence of using __slots__)
            state['set_hook']=param.identity_hook
    preprocess_state(param.Number,_add_set_hook)
    preprocess_state(R.BoundingRegionParameter,_add_set_hook)
     
support[12028] = Number_and_BoundingRegion_add_set_hook    


def renamed_sheetview_norm_factor():
    # CEBALERT: stuff to avoid overwriting existing dict could be extracted and made reusable to avoid
    # repeating it in all param_name_changes, param_moves
    sv_name_changes = PicklableClassAttributes.param_name_changes.get(
        'topo.base.sheetview.SheetView',{})
    sv_name_changes.update(
        {'norm_factor':'cyclic_range'})
    PicklableClassAttributes.param_name_changes['topo.base.sheetview.SheetView']=sv_name_changes

def moved_featuremaps_selectivity_multiplier():
    fm_moves = PicklableClassAttributes.param_moves.get(
        'topo.analysis.featureresponses.FeatureMaps',{})
    fm_moves.update(
        {'selectivity_multiplier':('topo.misc.distribution.DSF_WeightedAverage','selectivity_scale')})
    PicklableClassAttributes.param_moves['topo.analysis.featureresponses.FeatureMaps'] = fm_moves

def reorganized_analysis():
    renamed_sheetview_norm_factor()
    moved_featuremaps_selectivity_multiplier()
    
support[11904] = reorganized_analysis

def moved_picklableclassattributes():
    param.parameterized.PicklableClassAttributes = PicklableClassAttributes

support[12089] = moved_picklableclassattributes

######################################################################
######################################################################

# patches for external dependencies

# pickle stores path to modules; if path to module changes, module
# can't be imported, so snapshot can't be opened.

external_patches = {}

def numpy_core_defmatrix():
    import numpy.matrixlib.defmatrix
    allow_import(numpy.matrixlib.defmatrix,'numpy.core.defmatrix')


import numpy
if _version_greater_or_equal(numpy,decimal.Decimal("1.4")):
    external_patches['Support numpy.core.defmatrix for numpy>=1.4'] = numpy_core_defmatrix

# CB: About support for things like the numpy defmatrix change
# (above). When someone asks to load a snapshot, and they have
# numpy>=1.4, we register an import hook so that if the snapshot tries
# to import numpy.core.defmatrix, numpy.matrixlib.defmatrix will be
# provided. This happens without testing that the snapshot will
# actually try to import numpy.core.defmatrix (i.e. "import
# numpy.core.defmatrix" will work (with a warning) after loading *any*
# snapshot if version of numpy>=1.4). As far as I can see, the only
# way of improving this would be either (a) to set
# cPickle.Unpickler.find_global and make it handle all the various
# import changes (plus normal imports; there's no default because
# that's hidden in C) or (b) store version numbers of imported modules
# in the snapshot and use those to figure out what to do.



######################################################################
######################################################################

# CEBALERT: not sure whether to keep the "-l" option to Topographica.
# The idea is that we could start old scripts by loading legacy
# support, and then save a script_repr, thus allowing the script to be
# updated automatically. Not sure script_repr works well enough
# yet. (Should alter the -l option to take an optional release and svn
# version number.)
def install_legacy_support(release="0.9.7",version=None):
    if version is None:
        version = releases[release]

    assert version>=releases[release], "Release/version mismatch."

    SnapshotSupport.apply_external_patches()
    SnapshotSupport.apply_support(version)



"""
Unit test for Simulation
$Id$
"""

import unittest
import copy
import pickle

from numpy.oldnumeric import array
from topo.base.simulation import Simulation,EPConnection,EPConnectionEvent,Event
from topo.ep import *

from topo.base.cf import CFSheet, CFProjection

from utils import new_simulation

import topo

# CEBALERT: not a complete test of Simulation

class TestSimulation(unittest.TestCase):


    def test_register_is_true(self):
        sim1 = new_simulation(name="test_singleton")
        assert sim1 is topo.sim
        
        sid = id(topo.sim['S'])

        sim2 = copy.copy(topo.sim)
        assert sim2 is sim1

        sim3 = copy.deepcopy(topo.sim)        
        assert sim3 is sim1

        self.assertEqual(id(sim3['S']),sid)

        topo.sim['S'].precedence=111
        p = pickle.dumps(topo.sim,2)        
        topo.sim['S'].precedence=5

        Simulation.register=False # to ensure the object's own register is being used
        sim4 = pickle.loads(p)
        Simulation.register=True
        
        assert sim4 is sim1
        assert topo.sim['S'].precedence==111,"%s"%topo.sim['S'].precedence

        
        


    def test_register_is_false(self):
        sim1 = new_simulation(register=False)
        assert sim1 is not topo.sim
        
        sid = id(sim1['S'])

        sim2 = copy.copy(sim1)
        assert sim2 is not sim1
        assert sim2 is not topo.sim 
        
        sim3 = copy.deepcopy(sim1)        
        assert sim3 is not sim1
        assert sim3 is not topo.sim

        self.assertNotEqual(id(sim3['S']),sid)

        new_simulation(register=True)
        sim1['S'].precedence=111
        p = pickle.dumps(sim1,2)        
        topo.sim['S'].precedence=5
        sim4 = pickle.loads(p)

        assert sim4 is not sim1
        assert sim4 is not topo.sim
        assert topo.sim['S'].precedence==5


        
    
    def test_event_copy(self):
        """
        Test to make sure that EPConnectionEvent copies the underlying data
        on construction.
        """
        s = Simulation()
        data = array([4,3])
        epc = EPConnection()
        se = EPConnectionEvent(1,epc,data)
        se.data[0] = 5
        assert data[0] != se.data[0], 'Matrices should be different'
        se2 = copy.copy(se)
        assert se is not se2, 'Objects are the same'

    def test_state_stack(self):
        s = Simulation()
        s['pulse1'] = PulseGenerator(period = 1)
        s['pulse2'] = PulseGenerator(period = 3)
        s['sum_unit'] = SumUnit()
        s.connect('pulse1','sum_unit',delay=1)
        s.connect('pulse2','sum_unit',delay=1)
        s.run(1.0)
        s.state_push()
        self.assertEqual(len(s._events_stack),1)
        s.state_pop()
        self.assertEqual(len(s._events_stack),0)
        

    def test_event_cmp(self):

        e1 = Event(1)
        e1a = Event(1)
        e2 = Event(2)

        assert e1 is not e1a
        assert e1 ==  e1a
        assert e1 < e2
        assert e1 == e1
        assert e2 > e1
        assert e2 == e2
        assert e1 != e2
        assert e2 != e1
        
    def test_event_insert(self):
        s = Simulation()

        e1 = Event(1)
        e1a = Event(1)
        e2 = Event(2)
        e2a = Event(2)

        s.enqueue_event(e1)
        s.enqueue_event(e2)
        s.enqueue_event(e2a)
        s.enqueue_event(e1a)

        s.enqueue_event(Event(0))

        assert len(s.events) == 5, 'Event queue has %d events, should have 5.' % len(s.events)

        assert s.events[1] == e1
        assert s.events[2] == e1a
        assert s.events[3] == e2
        assert s.events[4] == e2a

        
    def test_get_objects(self):
        s = Simulation()

        s['pulse1'] = PulseGenerator(period = 1)
        s['pulse2'] = PulseGenerator(period = 3)
        s['sum_unit'] = SumUnit()
        n1 = s['pulse1'].name
        n2 = s['pulse2'].name

        s.connect('pulse1','sum_unit',delay=1)
        s.connect('pulse2','sum_unit',delay=1)
        t1 = s.objects()
        e1 = [ep for ep in t1.values() if isinstance(ep,PulseGenerator) and ep.name == n1]
        t2 = s.objects()
        e2 = [ep for ep in t2.values() if isinstance(ep,PulseGenerator) and ep.name == n2]
        assert e1.pop().name == n1, 'Object names do not match'
        assert e2.pop().name == n2, 'Object names do not match'


##     def test_garbage_collection(self):
##         """something"""
##         s = Simulation(register=False)
##         s['v1']=CFSheet(nominal_density=24)
##         s['v2']=CFSheet(nominal_density=24)
##         s.connect('v1','v2',name='Afferent',delay=0.05,connection_type=CFProjection)
##         from weakref import WeakValueDictionary as W
##         w=W()
##         w['v1'] = s['v1']
##         w['v2'] = s['v2']

##         # need to cause s to be deleted: s in this namespace should
##         # be only reference to s (apart from circular ones...)
##         del s

##         try:
##             w['v1']
##         except KeyError:
##             pass
##         else:
##             raise ValueError("v1 was not deleted")

##         try:
##             w['v2']
##         except KeyError:
##             pass
##         else:
##             raise ValueError("v2 was not deleted")


suite = unittest.TestSuite()
suite.addTest(unittest.makeSuite(TestSimulation))



"""
A family of function objects for transforming one set of coordinates into
another.

Coordinate mapper functions are useful for defining magnifications and
other kinds of transformations on sheet coordinates, e.g. for defining
retinal magnification using a CFProjection.  A CoordinateMapperFn
(e.g. MagnifyingMapper), is applied to an (x,y) pair and returns a new
(x,y) pair.  To apply a mapping to a CF projection, set the
CFProjection's coord_mapper parameter to an instance of the desired
CoordinateMapperFn.

$Id$
"""

from math import atan,pi,atan2

from numpy import exp,log,sqrt,sin,cos,ones,dot
from numpy.matlib import matrix

import param

from topo.base.functionfamily import CoordinateMapperFn
from topo.misc.util import signabs
from topo import numbergen




class ConstantMapper(CoordinateMapperFn):
    """
    Map all values to the same constant, pre-specified coordinates.
    """
    
    x_cons = param.Number(default=0.0,doc="""
       Constant x value returned by the mapping.""")
    
    y_cons = param.Number(default=0.0, doc="""
       Constant y value returned by the mapping.""")

    def __call__(self, x, y):
        # Ignores all (x,y), always returning (x_cons,y_cons)
        return self.x_cons, self.y_cons


class Pipeline(CoordinateMapperFn):
    """
    Applies a sequence of coordmappers, left to right.
    """
    
    mappers=param.List(default=[],
        doc="The sequence of mappers to apply.")

    def __call__(self,x,y):
        return reduce( lambda args,f: apply(f,args),
                       [(x,y)] + self.mappers )


class Jitter(CoordinateMapperFn):
    """
    Additively modifies calculated x,y coordinates, e.g. with random noise.
    """

    scale = param.Parameter(default=0.0,doc="Amount of jitter.")

    gen = param.Parameter(default=numbergen.UniformRandom(),doc=
        "Number generator to use, typically a random distribution.")

    def __call__(self,x,y):
        return x+(self.gen()-0.5)*self.scale,y+(self.gen()-0.5)*self.scale


class NormalJitter(CoordinateMapperFn):
    """
    Additively modifies calculated x,y coordinates with Gaussian random noise.
    """

    gen = param.Parameter(default=numbergen.NormalRandom(),doc=
         "Number generator to use, typically a Gaussian distribution.")
    
    def __call__(self,x,y):
        return x+self.gen(),y+self.gen()


class Grid(CoordinateMapperFn):
    """
    Divides the 2D area into a grid, where all points within each grid
    element map to the same location.
    """

    xdensity = param.Number(default=1, bounds=(0,None), doc="""
        Number of columns per 1.0 input sheet distance horizontally.""")

    ydensity = param.Number(default=1, bounds=(0,None), doc="""
        Number of rows per 1.0 input sheet distance vertically.""")

    def __call__(self,x,y):
        xd=self.xdensity
        yd=self.ydensity
        
        xquant=(1.0/xd)*(int(xd*(x+0.5))-(0.5*(xd-1)))
        yquant=(1.0/yd)*(int(yd*(y+0.5))-(0.5*(yd-1)))
        
        return  xquant,yquant


class Polar2Cartesian(CoordinateMapperFn):
    """
    Map from polar (radius,angle) to Cartesian (x,y) coordinates.
    """

    degrees=param.Boolean(default=True,
        doc="Indicates whether the input angle is in degrees or radians.")

    def __call__(self, r, theta):

        if self.degrees:
            theta = theta * pi/180

            return r*cos(theta), r*sin(theta)
                          

class Cartesian2Polar(CoordinateMapperFn):
    """
    Maps from Cartesian (x,y) to polar (radius,angle).
    """

    degrees = param.Boolean(default=True,
        doc="Indicates whether the output angle is in degrees or radians.")

    negative_radii = param.Boolean(default = False,
        doc="""If true, coordinates with negative x values will be
        given negative radii, and angles between -90 and 90
        degrees. (useful for mapping to saccade amplitude/direction space)""")

    def __call__(self, x, y):

        if self.negative_radii:
            xsgn,xabs = signabs(x)            
            radius = xsgn * sqrt(x*x+y*y)
            angle = atan2(y,xabs)
        else:
            radius = sqrt(x*x+y*y)
            angle = atan2(y,x)

        if self.degrees:
            angle *= 180/pi

        return radius,angle
                          



class AffineTransform(CoordinateMapperFn):
    """
    Remaps the input with an affine transform.

    This mapper allows the specification of an arbitrary combination
    of translation, rotation and scaling via a transform
    matrix. Single translations, etc, can be specified more simply
    with the subclasses Translate2d, Rotate2d, and Scale2d.
    """
    
    matrix = param.Parameter(default=ones((3,3)),doc="""
       The affine transformation matrix.  The functions
       Translate2dMat, Rotate2dMat, and Scale2dMat generate affine
       transform matrices that can be multiplied together to create
       combination transforms.  E.g. the matrix::

         Translate2dMat(3,0)*Rotate2d(pi/2)

       will shift points to the right by 3 units and rotate them around
       the origin by 90 degrees.""")

    def __init__(self, **kw):
        super(AffineTransform,self).__init__(**kw)

        # This buffer prevents having to allocate memory for each point.
        self._op_buf = matrix([[0.0],
                               [0.0],
                               [1.0]])
        
    def __call__(self, x, y):

        ## JPHACKALERT: If the coordmapper interface took a matrix of
        ## x/y column vectors, instead of x and y separately, affine
        ## transforms could be applied to all the points in a single
        ## matrix operation.  This would probably require revision of
        ## some of the other coordmapper functions, but it might allow
        ## some optimization in, e.g., CFProjection intialization by
        ## allowing all the CF positions to be computed at once.

        ## JPALERT: This is the easy way, but it allocates a matrix
        ## for the result.  It might be faster to unroll the
        ## computation.

        self._op_buf[0] = x
        self._op_buf[1] = y
        result = dot(self.matrix,self._op_buf)
               
        return result[0,0],result[1,0]


def Translate2dMat(xoff,yoff):
    """
    Return an affine transformation matrix that translates points by
    the offset (xoff,yoff).
    """
    return matrix([[1, 0, xoff],
                   [0, 1, yoff],
                   [0, 0,   1 ]])


def Rotate2dMat(t):
    """
    Return an affine transformation matrix that rotates the points
    around the origin by t radians.
    """
    return matrix([[cos(t), -sin(t), 0],
                   [sin(t),  cos(t), 0],
                   [  0   ,    0   , 1]])


def Scale2dMat(sx,sy):
    """
    Return an affine translation matrix that scales the points
    toward/away from the origin by a factor of sx on x-axis and sy on
    the y-axis.
    """
    return matrix([[sx,  0, 0],
                   [ 0, sy, 0],
                   [ 0,  0, 1]])


class Translate2d(AffineTransform):
    """
    Translate the input by xoff,yoff.
    """
    xoff = param.Number(default=0.0)
    yoff = param.Number(default=0.0)

    def __init__(self,**kw):
        super(Translate2d,self).__init__(**kw)
        self.matrix = Translate2dMat(self.xoff,self.yoff)


class Rotate2d(AffineTransform):
    """
    Rotate the input around the origin by an angle in radians.
    """
    angle = param.Number(default=0.0)

    def __init__(self,**kw):
        super(Rotate2d,self).__init__(**kw)
        self.matrix = Rotate2dMat(self.angle)


class Scale2d(AffineTransform):
    """
    Scale the input along the x and y axes by sx and sy, respectively.
    """
    sx = param.Number(default=1.0)
    sy = param.Number(default=1.0)

    def __init__(self,**kw):
        super(Scale2d,self).__init__(**kw)
        self.matrix = Scale2dMat(self.sx,self.sy)


class SingleDimensionMapper(CoordinateMapperFn):
    """
    Coordinate Mapper that uses an origin-centered 1-D mapping function.

    An abstract mapping function for coordinate mappers that remap
    based on the radius, x, or y individually. Subclasses should override
    _map_fn(self,z). 
    """
    __abstract = True

    in_range = param.Number(default=0.5*sqrt(2),bounds=(0,None),doc="""
       The maximum range of the mapping input.""")
    out_range = param.Number(default=0.5*sqrt(2),bounds=(0,None), doc="""
       The maximum range of the output.""")
    remap_dimension = param.ObjectSelector(default='radius',
        objects=['radius','x','y','xy'],doc="""
        The dimension to remap. ('xy' remaps x and y independently.)""")


    def __call__(self,x,y):

        if self.remap_dimension == 'radius':
            r = sqrt(x**2 + y**2)
            a = atan2(x,y)
            new_r = self._map_fn(r)
            xout = new_r * sin(a)
            yout = new_r * cos(a)
        else:
            if 'x' in self.remap_dimension:
                xout = self._map_fn(x)
            else:
                xout = x

            if 'y' in self.remap_dimension:
                yout = self._map_fn(y)
            else:
                yout = y

        return xout,yout

    def _map_fn(self,z):
        raise NotImplementedError


class MagnifyingMapper(SingleDimensionMapper):
    """
    Exponential (magnifying) mapping function.

    Provides a mapping that magnifies the center of the activity image. 
    Parameter k indicates amount of magnification, where 0 means no
    magnification.
    """
    
    k = param.Number(default=1.0,bounds=(0,None))

    def _map_fn(self,z):
        k = self.k
        if k == 0:
            return z
        else:
            sgn,z = signabs(z)
            return sgn * self.out_range * (exp(z/self.in_range*k)-1)/(exp(k)-1)


class ReducingMapper(SingleDimensionMapper):
    """
    Provides a mapping that reduces the center of the activity.

    Roughly the inverse of MagnifyingMapper.  k indicates amount of reduction.    
    """
    k = param.Number(default=1.0,bounds=(0,None))
    
    def _map_fn(self,z):
        k = self.k
        sgn,z = signabs(z)
        return sgn * self.out_range * log(z/self.in_range*k+1)/log(k+1)
        

class OttesSCMapper(CoordinateMapperFn):
    """
    Abstract class for superior colliculus mappings.

    Subclasses of this class implement afferent and efferent mappings
    from Ottes et al. (1986) Vision Research 26:857-873.
    
    Default constant values are from Table 1, ibid.  
    """
    __abstract = True 

    
    A = param.Number(default=5.3, doc="""
       Shape parameter A, in degrees""")
    Bu = param.Number(default=1.8, doc="""
       Rostral-caudal scale parameter, in mm""")
    Bv = param.Number(default=1.8, doc="""
       Orthogonal (medial-lateral?) scale paraemter in mm/deg""")

    mm_scale = param.Number(default=8.0,doc="""
       Scale factor to convert constants Bu and Bv from mm to sheet
       units.  Expressed in mm/unit. """)

    amplitude_scale = param.Number(default=1,doc="""
        Scale factor for saccade command amplitude, expressed in
        degrees per unit of sheet.  Indicates how large a
        saccade is represented by the x-component of the command
        input.""")
    
    direction_scale = param.Number(default=1,doc="""
        Scale factor for saccade command direction, expressed in
        degrees per unit of sheet.  Indicates what direction of saccade
        is represented by the y-component of the command input.""")
    

    def __call__(self,x,y):
        raise NotImplementedError        


class OttesSCMotorMapper(OttesSCMapper):
    """
    Efferent superior colliculus mapping.

    Provides the output/motor mapping from SC as defined by Ottes et al.
    (see superclass docs for reference)

    The mapping allows the creation of a single sheet representing
    both colliculi, one in the x-positive hemisheet and the other in
    the x-negative hemisheet.
    """
    def __call__(self,x,y):

        A = self.A 
        Bu = self.Bu / self.mm_scale
        Bv = self.Bv / self.mm_scale 

        R = x * self.amplitude_scale 
        phi = y * self.direction_scale

        Rsign,R = signabs(R)

        u,v = ottes_mapping(R,phi,A,Bu,Bv)
        return Rsign*u,v

        

class OttesSCSenseMapper(OttesSCMapper):
    """
    Afferent superior colliculus mapping.

    Provides the input/retinal mapping to SC as defined by Ottes et al.
    (see superclass docs for reference)

    The mapping allows the creation of a single sheet representing
    both colliculi, one in the x-positive hemisheet and the other in
    the x-negative hemisheet.

    [NOTE: see warning in docs for ottes_inverse_mapping().]
    """

    def __call__(self,x,y):

        A  = self.A 
        Bu = self.Bu 
        Bv = self.Bv 

        u = x * self.mm_scale
        v = y * self.mm_scale

        usgn,u = signabs(u)

        R,phi = ottes_inverse_mapping(u,v,A,Bu,Bv)

        return (usgn*R/self.amplitude_scale,
                phi/self.direction_scale)


def ottes_mapping(R,phi,A,Bu,Bv):
    """
    The efferent mapping function from Ottes et al. (1986)
    Vision Research 26:857-873.

    Takes saccade with amplitude R (in degrees) and direction
    phi (in degrees), and returns a location u,v on the colliculus
    in mm, where the u axis is rostral/caudal, and the v axis is
    medial/lateral.
    """
    
    phi *= pi/180
    u = Bu * (log(sqrt(R**2 + A**2 + 2*A*R*cos(phi))) - log(A))
    v = Bv * atan((R*sin(phi))/(R*cos(phi)+A))        
    return u,v


def ottes_inverse_mapping(u,v,A,Bu,Bv):
    """
    Approximate inverse of ottes_mapping(), using the inverse function
    provided in the appendix of Ottes et al. (1986) Vision Research 26:857-873

    Takes takes a location u,v on the colliculus in mm and maps
    to a retinal eccentricity R and direction phi, in degrees.
    Inverse is approximate, with increasing error as positions near the
    edges of the collicular sheet. (I.e. with high absolute v value).
    """
    
    rads = pi/180
    R   = A * sqrt(exp(2*u/Bu) - 2*exp(u/Bu)*cos(rads*v/Bv) + 1)
    #phi = atan( (exp(u/Bu)*sin(rads*v/Bv)) / (exp(u/Bu)*cos(rads*v/Bv) -1) )
    phi = atan2( (exp(u/Bu)*sin(rads*v/Bv)), (exp(u/Bu)*cos(rads*v/Bv) -1) ) * 180/pi

    # JPALERT: Don't know why we have to multiply by 180/pi twice, but the answers
    # are way off without it.  Is the bug in my code, or in the original formula?
    return R,phi*180/pi
                 

# JPALERT: Temporary testing function.  Will disappear eventually.
def test_ottes_inverse():

    A,Bu,Bv = 5.3,1.8,1.8

    print '%10s %10s | %10s %10s | %10s %10s | %s' \
          % ('R in','phi in','R out','phi out','R err','phi err','phi ratio')
    for r in range(10,60,10):
        for phi in range(-60,60,10):
            if  phi != 0:
                u,v = ottes_mapping(r,phi,A,Bu,Bv)
                r2,phi2 = ottes_inverse_mapping(u,v,A,Bu,Bv)
                print '%10.2f %10.2f | %10.2f %10.2f | %10.2f %10.2f | %.2f' \
                      % (r,phi,r2,phi2,r2-r,phi2-phi,phi/phi2)


_public = list(set([_k for _k,_v in locals().items()
                    if isinstance(_v,type) \
                    and issubclass(_v,CoordinateMapperFn)]))

# Automatically discover all .py files in this directory.
import os,fnmatch
__all__ = _public + [f.split('.py')[0] for f in os.listdir(__path__[0]) if fnmatch.fnmatch(f,'[!._]*.py')]
del f,os,fnmatch



"""
$Id$
"""

### JCALERT: Does this file is of any interrest now?


import unittest

class TestDummy(unittest.TestCase):
    def test_add(self):
        self.assertEqual(1+1,2)



suite = unittest.TestSuite()
#  Uncomment the following line of code, to disable the test if
#  $DISPLAY is undefined.  Used mainly for GUI testing.
# suite.requires_display = True
suite.addTest(unittest.makeSuite(TestDummy))



"""
Neural sheet objects and associated functions.

The Sheet class is the base class for EventProcessors that simulate
topographically mapped sheets of units (neurons or columns).  A Sheet
is an EventProcessor that maintains a rectangular array of activity
values, and sends the contents of this array as the data element in
events.

The Sheet itself is a SheetCoordinateSystem, so that units may be
accessed by sheet or matrix coordinates. In general, however, sheets
should be thought of as having arbitrary density and sheet coordinates
should be used wherever possible, except when the code needs direct
access to a specific unit.  By adhering to this convention, one should
be able to write and debug a simulation at a low density, and then
scale it up to run at higher densities (or down for lower densities)
simply by changing e.g. Sheet.nominal_density.

$Id$
"""



from numpy import zeros,array,arange,meshgrid
from numpy import float64

import param

from simulation import EventProcessor
from sheetcoords import SheetCoordinateSystem
from boundingregion import BoundingBox, BoundingRegionParameter
from functionfamily import TransferFn


activity_type = float64

# (disable W0223 because input_event is deliberately still not implemented)
class Sheet(EventProcessor,SheetCoordinateSystem):  # pylint: disable-msg=W0223
    """
    The generic base class for neural sheets.

    See SheetCoordinateSystem for how Sheet represents space, and
    EventProcessor for how Sheet handles time.

    output_fns are functions that take an activity matrix and produce
    an identically shaped output matrix. The default is having no
    output_fns.
    """
    __abstract = True

    nominal_bounds = BoundingRegionParameter(BoundingBox(radius=0.5),constant=True,doc="""
            User-specified BoundingBox of the Sheet coordinate area
            covered by this Sheet.  The left and right bounds--if
            specified--will always be observed, but the top and bottom
            bounds may be adjusted to ensure the density in the y
            direction is the same as the density in the x direction.
            In such a case, the top and bottom bounds are adjusted
            so that the center y point remains the same, and each
            bound is as close as possible to its specified value. The
            actual value of this Parameter is not adjusted, but the
            true bounds may be found from the 'bounds' attribute
            of this object.
            """)
    
    nominal_density = param.Number(default=10,constant=True,doc="""
            User-specified number of processing units per 1.0 distance
            horizontally or vertically in Sheet coordinates. The actual
            number may be different because of discretization; the matrix
            needs to tile the plane exactly, and for that to work the
            density might need to be adjusted.  For instance, an area of 3x2
            cannot have a density of 2 in each direction. The true density
            may be obtained from either the xdensity or ydensity attribute
            (since these are identical for a Sheet).
            """)
    
    plastic = param.Boolean(True,doc="""
            Setting this to False tells the Sheet not to change its
            permanent state (e.g. any connection weights) based on
            incoming events.
            """)

    precedence = param.Number(default = 0.1, softbounds=(0.0,1.0),doc="""
            Allows a sorting order for Sheets, e.g. in the GUI.""")

    row_precedence = param.Number(default = 0.5, softbounds=(0.0,1.0),doc="""
            Allows grouping of Sheets before sorting precedence is
            applied, e.g. for two-dimensional plots in the GUI.""")

    layout_location = param.NumericTuple(default = (-1,-1),precedence=-1,doc="""
            Location for this Sheet in an arbitrary pixel-based space
            in which Sheets can be laid out for visualization.""")

    output_fns = param.HookList(default=[],class_=TransferFn,
        doc="Output function(s) to apply (if apply_output_fns is true) to this Sheet's activity.")

    apply_output_fns=param.Boolean(default=True,
        doc="Whether to apply the output_fn after computing an Activity matrix.")


    def _get_density(self):
        return self.xdensity

    density = property(_get_density,doc="""The sheet's true density (i.e. the xdensity, which is equal to the ydensity for a Sheet.)""")

    def __init__(self,**params):
        """
        Initialize this object as an EventProcessor, then also as
        a SheetCoordinateSystem with equal xdensity and ydensity.

        sheet_views is a dictionary that stores SheetViews,
        i.e. representations of the sheet for use by analysis or plotting
        code.    
        """
        EventProcessor.__init__(self,**params)

        # Initialize this object as a SheetCoordinateSystem, with
        # the same density along y as along x.
        SheetCoordinateSystem.__init__(self,self.nominal_bounds,self.nominal_density)

        n_units = round((self.lbrt[2]-self.lbrt[0])*self.xdensity,0)
        if n_units<1: raise ValueError(
           "Sheet bounds and density must be specified such that the "+ \
           "sheet has at least one unit in each direction; " \
           +self.name+ " does not.")

        # setup the activity matrix
        self.activity = zeros(self.shape,activity_type)

        # For non-plastic inputs
        self.__saved_activity = []
        self._plasticity_setting_stack = []  
        self.sheet_views = {}        


    ### JABALERT: This should be deleted now that sheet_views is public
    ### JC: shouldn't we keep that, or at least write a function in
    ### utils that deletes a value in a dictinnary without returning an
    ### error if the key is not in the dict?  I leave for the moment,
    ### and have to ask Jim to advise.
    def release_sheet_view(self,view_name):
        """
        Delete the dictionary entry with key entry 'view_name' to save
        memory.
        """
        if self.sheet_views.has_key(view_name):   
            del self.sheet_views[view_name]


    # CB: what to call this? sheetcoords()? sheetcoords_of_grid()? idxsheetcoords()?
    def sheetcoords_of_idx_grid(self):
        """
        Return an array of x-coordinates and an array of y-coordinates
        corresponding to the activity matrix of the sheet.
        """
        nrows,ncols=self.activity.shape

        C,R = meshgrid(arange(ncols),
                       arange(nrows))

        X,Y = self.matrixidx2sheet(R,C)
        return X,Y
        

    # CB: check whether we need this function any more.
    def row_col_sheetcoords(self):
        """
        Return an array of Y-coordinates corresponding to the rows of
        the activity matrix of the sheet, and an array of
        X-coordinates corresponding to the columns.
        """
        # The row and column centers are returned in matrix (not
        # sheet) order (hence the reversals below).
        nrows,ncols = self.activity.shape
        return self.matrixidx2sheet(arange(nrows-1,-1,-1),arange(ncols))[::-1]


    # CBALERT: to be removed once other code uses
    # row_col_sheetcoords() or sheetcoords_of_idx_grid().
    def sheet_rows(self):
        return self.row_col_sheetcoords()[0]
    def sheet_cols(self):
        return self.row_col_sheetcoords()[1]


    # CEBALERT: haven't really thought about what to put in this. The
    # way it is now, subclasses could make a super.activate() call to
    # avoid repeating some stuff.
    def activate(self):
        """
        Collect activity from each projection, combine it to calculate
        the activity for this sheet, and send the result out. 

        Subclasses will need to override this method to whatever it
        means to calculate activity in that subclass.
        """        
        if self.apply_output_fns:
            for of in self.output_fns:
                of(self.activity)
    
        self.send_output(src_port='Activity',data=self.activity)


    def state_push(self):
        """
        Save the current state of this sheet to an internal stack.

        This method is used by operations that need to test the
        response of the sheet without permanently altering its state,
        e.g. for measuring maps or probing the current behavior
        non-invasively.  By default, only the activity pattern of this
        sheet is saved, but subclasses should add saving for any
        additional state that they maintain, or strange bugs are
        likely to occur.  The state can be restored using state_pop().

        Note that Sheets that do learning need not save the
        values of all connection weights, if any, because
        plasticity can be turned off explicitly.  Thus this method
        is intended only for shorter-term state.
        """
        self.__saved_activity.append(array(self.activity))
        EventProcessor.state_push(self)
        for of in self.output_fns:
            if hasattr(of,'state_push'):
                of.state_push()


    def state_pop(self):
        """
        Pop the most recently saved state off the stack.

        See state_push() for more details.
        """
        self.activity = self.__saved_activity.pop()
        EventProcessor.state_pop(self)
        for of in self.output_fns:
            if hasattr(of,'state_pop'):
                of.state_pop()


    def activity_len(self):
        """Return the number of items that have been saved by state_push()."""
        return len(self.__saved_activity)


    def override_plasticity_state(self, new_plasticity_state):
        """
        Temporarily override plasticity of medium and long term internal state.

        This function should be implemented by all subclasses so that
        it preserves the ability of the Sheet to compute activity,
        i.e. to operate over a short time scale, while preventing any
        lasting changes to the state (if new_plasticity_state=False).

        Any operation that does not have any lasting state, such as
        those affecting only the current activity level, should not
        be affected by this call.

        By default, simply saves a copy of the plastic flag to an
        internal stack (so that it can be restored by
        restore_plasticity_state()), and then sets plastic to
        new_plasticity_state.
        """
        self._plasticity_setting_stack.append(self.plastic)
        self.plastic=new_plasticity_state


    def restore_plasticity_state(self):
        """
        Restores plasticity of medium and long term internal state after
        a override_plasticity_state call.

        This function should be implemented by all subclasses to
        remove the effect of the most recent override_plasticity_state call,
        i.e. to restore plasticity of any type that was overridden.
        """
        self.plastic = self._plasticity_setting_stack.pop()


    def n_bytes(self):
        """
        Return a lower bound for the memory taken by this sheet, in bytes.

        Typically, this number will include the activity array and any
        similar arrays, plus any other significant data owned (in some
        sense) by this Sheet.  It will not usually include memory
        taken by the Python dictionary or various "housekeeping"
        attributes, which usually contribute only a small amount to
        the memory requirements.

        Subclasses should reimplement this method if they store a
        significant amount of data other than in the activity array.
        """
        return self.activity.nbytes




"""
Topographica-specific changes to the standard pydoc command.

Moves generated pydoc HTML files to the docs directory after they are
created by pydoc.  Intended for use with MAKE, from within the base
topographica/ directory.

The generated index.html file is for the topo/__init__.py file, which
does not necessarily catalog every .py file in topo/.  To see all
files, select the 'index' link at the top of one of the html docs.

To generate documentation, enter 'make docs' from the base
topographica/ directory, which will call this file on the Topographica
sources.

From the Makefile (Tabs have been stripped)::

    cleandocs:
        - rm -r docs
        
    docs: topo/*.py
        mkdir -p docs
        ./topographica topo/gendocs.py
        mv docs/topo.__init__.html docs/index.html

$Id$
"""


import pydoc, glob, os
from os.path import isdir
from re import search, sub
from copy import copy


TOPO = 'topo'   # Subdirectory with Topographica source
DOCS = 'doc/Reference_Manual'   # Subdirectory to place Docs
pydoc.writing = 1


def _file_list(base_name):
    """
    Recursively generate a list of files and directories to document.

    base_name is the relative path to add to directories and files.
    Files that match CVS or __init__.py are ignored.
    """
    filelist = [f for f in glob.glob(base_name + '/*')
                if (isdir(f) or search('.py$',f))
                and not search('CVS',f) and not search('__init__.py',f)]
    for f in copy(filelist):
        if isdir(f):
            filelist = filelist + _file_list(f)
    return filelist


def filename_to_docname(f):
    """
    Convert a path name into the PyDoc filename it will turn into.

    If the path name ends in a .py, then it is cut off.  If there is no
    extension, the name is assumed to be a directory.
    """
    f = sub('/','.',f)
    if search('.py$',f):
        f = sub('.py$','.html',f)
    else:
        f = f + '.html'
    return f


def filename_to_packagename(f):
    """
    Convert a path name into the Python dotted-notation package name.

    If the name ends in a .py, then cut it off.  If there is no
    extension, the name is assumed to be a directory, and nothing is done
    other than to replace the '/' with '.'
    """
    f = sub('/','.',f)
    if search('.py$',f):
        f = sub('.py$','',f)
    return f

    
def generate_docs():
    """
    Generate all pydoc documentation files within a docs directory under
    ./topographica according to the constant DOCS.  After generation,
    there is an index.html that displays all the modules.  Note that
    if the documentation is being changed, it may be necessary to call
    'make cleandocs' to force a regeneration of documentation.  (We
    don't want to regenerate all the documentation each time a source
    file is changed.)
    """
    # os.system('rm -rf ' + DOCS + '/*') # Force regeneration
    filelist =  _file_list(TOPO) + [TOPO]
    for i in filelist:
        f = filename_to_docname(i)
        if not glob.glob(DOCS + '/' + f):
            pydoc.writedoc(filename_to_packagename(i))
        if glob.glob(f):
            cline = 'mv -f ' + f + ' ' + DOCS + '/'
            os.system(cline)
        else:                   
            filelist.remove(i)  
        

if __name__ == '__main__':
    generate_docs()



"""
Topographica Bitmap Class.

Encapsulates the PIL Image class so that an input matrix can be displayed
as a bitmap image without needing to know about PIL proper.

There are three different base image Classes which inherit Bitmap:

PaletteBitmap  - 1 2D Matrix, 1 1D Color Map
HSVBitmap    - 3 2D Matrices, Color (H), Confidence (S), Strength (V)
RGBBitmap    - 3 2D Matrices, Red, Green, Blue Channels.

All maps are assumed to be on a nominal range of 0.0 to 1.0.  Matrices
are passed in as part of the constructor and the image is generaed.
For more information, see the documentation for each of the Bitmap
classes.

The encapsulated PIL Image is accessible through the .bitmap attribute.

$Id$
"""

import os
import Image
import ImageDraw
import ImageFont
from colorsys import hsv_to_rgb
import numpy.oldnumeric as Numeric
import numpy

import param
from param import resolve_path


# CEBALERT: can we just use load_default()? Do we even need TITLE_FONT
# at all?
try:
   import matplotlib
   _vera_path = resolve_path(os.path.join(matplotlib.__file__,'matplotlib/mpl-data/fonts/ttf/Vera.ttf'))
   TITLE_FONT = ImageFont.truetype(_vera_path,20)
except:
   TITLE_FONT = ImageFont.load_default()


### JCALERT: To do:
###        - Update the test file.
###        - Write PaletteBitmap when the Palette class is fixed
###        - Get rid of accessing function (copy, show...) (should we really?)


class Bitmap(param.Parameterized):
    """
    Wrapper class for the PIL Image class.

    The main purpose for this base class is to provide a consistent
    interface for defining bitmaps constructed in various different
    ways.  The resulting bitmap is a PIL Image object that can be
    accessed using the normal PIL interface.

    If subclasses use the _arrayToImage() function provided, any
    pixels larger than the maximum that can be displayed will
    be counted before they are clipped; these are stored in the
    clipped_pixels attribute.
    """
    clipped_pixels = 0
    
    def __init__(self,image):
        self.image = image


    def __copy__(self):
        # avoid calling __getstate__ for copy (not required)
        image = self.image.copy()
        return Bitmap(image)

    # CB: could define a __deepcopy__ too, but we don't need
    # deepcopy to be fast.

    def __getstate__(self):
        """
        Return the object's state (as in the superclass), but replace
        the 'image' attribute's Image with a string representation.
        """
        state = super(Bitmap,self).__getstate__()
        import StringIO
        f = StringIO.StringIO()
        image = state['image']
        image.save(f,format=image.format or 'TIFF') # format could be None (we should probably just not save in that case)
        state['image'] = f.getvalue()
        f.close()

        return state

    def __setstate__(self,state):
        """
        Load the object's state (as in the superclass), but replace
        the 'image' string with an actual Image object.
        """
        import StringIO
        state['image'] = Image.open(StringIO.StringIO(state['image']))
        super(Bitmap,self).__setstate__(state)
        

    def show(self):
        """
        Renaming of Image.show() for the Bitmap.bitmap attribute.
        """
        self.image.show()

    def width(self): return self.image.size[0]
    def height(self): return self.image.size[1]

    def zoom(self, factor):
        """
        Return a resized Image object, given the input 'factor'
        parameter.  1.0 is the same size, 2.0 is doubling the height
        and width, 0.5 is 1/2 the original size.  The original Image
        is not changed.
        """
        if factor%1==0:
            # CEBALERT: work around PIL bug (see SF #2820821) so that
            # integer scaling works in the typical case (where an
            # image is being enlarged).
            a = numpy.array(self.image).repeat(int(factor),axis=0).repeat(int(factor),axis=1)
            zoomed = Image.fromarray(a,mode=self.image.mode)
        else:
            x,y = self.image.size
            zx, zy = int(x*factor), int(y*factor)
            zoomed = self.image.resize((zx,zy))
            
        return zoomed


    # CEBALERT: Might be worthwhile simplifying this
    # method. E.g. Image.fromarray() now exists (PIL>=1.1.6), and
    # probably various numpy operations can now be written more
    # clearly, too.
    def _arrayToImage(self, inArray):
        """
        Generate a 1-channel PIL Image from an array of values from 0 to 1.0.

        Values larger than 1.0 are clipped, after adding them to the total
        clipped_pixels.  Returns a one-channel (monochrome) Image.
        """
        
        # PIL 'L' Images use a range of 0 to 255, so we scale the
        # input array to match.  The pixels are scaled by 255, not
        # 256, so that 1.0 maps to fully white.
        max_pixel_value=255
        inArray = (Numeric.floor(inArray * max_pixel_value)).astype(Numeric.Int)

        # Clip any values that are still larger than max_pixel_value
        to_clip = (Numeric.greater(inArray.ravel(),max_pixel_value)).sum()
        if (to_clip>0):
            # CEBALERT: no explanation of why clipped pixel count is
            # being accumulated.
            self.clipped_pixels = self.clipped_pixels + to_clip
            inArray.clip(0,max_pixel_value,out=inArray)
            self.verbose("Bitmap: clipped",to_clip,"image pixels that were out of range")

        r,c = inArray.shape
        # The size is (width,height), so we swap r and c:
        newImage = Image.new('L',(c,r),None)
        newImage.putdata(inArray.ravel())
        return newImage
        

class PaletteBitmap(Bitmap):
    """
    Bitmap constructed using a single 2D array.
 
    The image is monochrome by default, but more colorful images can
    be constructed by specifying a Palette.
    """

    def __init__(self,inArray,palette=None):
        """
        inArray should have values in the range from 0.0 to 1.0.
        
        Palette can be any color scale depending on the type of ColorMap
        desired.  Examples:
        
        [0,0,0 ... 255,255,255] = grayscale
        [0,0,0 ... 255,0,0] = grayscale but through a Red filter.

        The default palette is grayscale, with 0.0 mapping to black
        and 1.0 mapping to white.
        """
        ### JABALERT: Should accept a Palette class, not a data
        ### structure, unless for some reason we want to get rid of
        ### the Palette classes and always use data structures
        ### instead.
        ### JC: not yet properly implemented anyway.
        max_pixel_value=255

        newImage = self._arrayToImage(inArray)
        if palette == None:
            palette = [i for i in range(max_pixel_value+1) for j in range(3)]
        newImage.putpalette(palette)
        newImage = newImage.convert('P')
        super(PaletteBitmap,self).__init__(newImage)



class HSVBitmap(Bitmap):
    """
    Bitmap constructed from 3 2D arrays, for hue, saturation, and value.

    The hue matrix determines the pixel colors.  The saturation matrix
    determines how strongly the pixels are saturated for each hue,
    i.e. how colorful the pixels appear.  The value matrix determines
    how bright each pixel is.

    An RGB image is constructed from the HSV matrices using
    hsv_to_rgb; the resulting image is of the same type that is
    constructed by RGBBitmap, and can be used in the same way.
    """

    def __init__(self,hue,sat,val):
        """Each matrix must be the same size, with values in the range 0.0 to 1.0."""
        shape = hue.shape # Assumed same as sat.shape and val.shape
        rmat = Numeric.zeros(shape,Numeric.Float)
        gmat = Numeric.zeros(shape,Numeric.Float)
        bmat = Numeric.zeros(shape,Numeric.Float)

        # Note: should someday file a feature request for PIL for them
        # to accept an image of type 'HSV', so that they will do this
        # conversion themselves, without us needing an explicit loop
        # here.  That should speed this up.
        ch = hue.clip(0.0,1.0)
        cs = sat.clip(0.0,1.0)
        cv = val.clip(0.0,1.0)
        
        for i in range(shape[0]):
            for j in range(shape[1]):
                r,g,b = hsv_to_rgb(ch[i,j],cs[i,j],cv[i,j])
                rmat[i,j] = r
                gmat[i,j] = g
                bmat[i,j] = b
        
        rImage = self._arrayToImage(rmat)
        gImage = self._arrayToImage(gmat)
        bImage = self._arrayToImage(bmat)

        super(HSVBitmap,self).__init__(Image.merge('RGB',(rImage,gImage,bImage)))



class RGBBitmap(Bitmap):
    """
    Bitmap constructed from three 2D arrays, for red, green, and blue.

    Each matrix is used as the corresponding channel of an RGB image.
    """

    def __init__(self,rMapArray,gMapArray,bMapArray):
        """Each matrix must be the same size, with values in the range 0.0 to 1.0."""
        rImage = self._arrayToImage(rMapArray)
        gImage = self._arrayToImage(gMapArray)
        bImage = self._arrayToImage(bMapArray)

        super(RGBBitmap,self).__init__(Image.merge('RGB',(rImage,gImage,bImage)))



class MontageBitmap(Bitmap):
    """
    A bitmap composed of tiles containing other bitmaps.

    Bitmaps are scaled to fit in the given tile size, and tiled
    right-to-left, top-to-bottom into the given number of rows and columns.
    """
    bitmaps = param.List(class_=Bitmap,doc="""
        The list of bitmaps to compose.""")

    rows = param.Integer(default=2, doc="""
        The number of rows in the montage.""")
    cols = param.Integer(default=2, doc="""
        The number of columns in the montage.""")    
    shape = param.Composite(attribs=['rows','cols'], doc="""
        The shape of the montage. Same as (self.rows,self.cols).""")

    margin = param.Integer(default=5,doc="""
        The size in pixels of the margin to put around each
        tile in the montage.""")

    tile_size = param.NumericTuple(default=(100,100), doc="""
        The size in pixels of a tile in the montage.""")

    titles = param.List(class_=str, default=[], doc="""
        A list of titles to overlay on the tiles.""")

    title_pos = param.NumericTuple(default=(10,10), doc="""
        The position of the upper left corner of the title in each tile.""")

    title_options = param.Dict(default={}, doc="""
        Dictionary of options for drawing the titles.  Dict should
        contain keyword options for the PIL draw.text method.  Possible
        options include 'fill' (fill color), 'outline' (outline color),
        and 'font' (an ImageFont font instance).  The PIL defaults will
        be used for any omitted options.""",
        instantiate=False)

    hooks = param.List(default=[], doc="""
        A list of functions, one per tile, that take a PIL image as
        input and return a PIL image as output.  The hooks are applied
        to the tile images before resizing.  The value None can be
        inserted as a placeholder where no hook function is needed.""")

    resize_filter = param.Integer(default=Image.NEAREST,doc="""
       The filter used for resizing the images.  Defaults
       to NEAREST.  See PIL Image module documentation for other
       options and their meanings.""")
    
    bg_color = param.NumericTuple(default=(0,0,0), doc="""
       The background color for the montage, as (r,g,b).""")

    def __init__(self,**params):
        ## JPALERT: The Bitmap class is a Parameterized object,but its
        ## __init__ doesn't take **params and doesn't call super.__init__,
        ## so we have to skip it.
        ## JAB: Good point; Bitmap should be modified to be more like
        ## other PO classes.
        param.Parameterized.__init__(self,**params)

        rows,cols = self.shape
        tilew,tileh = self.tile_size
        bgr,bgg,bgb = self.bg_color

        width  = tilew*cols + self.margin*(cols*2)
        height = tileh*rows + self.margin*(rows*2) 
        self.image = Image.new('RGB',(width,height),
                               (bgr*255,bgg*255,bgb*255))

        self.title_options.setdefault('font',TITLE_FONT)
        
        for r in xrange(rows):
            for c in xrange(cols):
                i = r*self.cols+c
                if i < len(self.bitmaps):
                    bm = self.bitmaps[i]
                    bmw,bmh = bm.image.size
                    if bmw > bmh:
                        bmh = int( float(tilew)/bmw * bmh )
                        bmw = tilew
                    else:
                        bmw = int( float(tileh)/bmh * bmw )
                        bmh = tileh

                    if self.hooks and self.hooks[i]:
                        f = self.hooks[i]
                    else:
                        f = lambda x:x
                    new_bm = Bitmap(f(bm.image).resize((bmw,bmh)))
                    if self.titles:
                        draw = ImageDraw.Draw(new_bm.image)
                        draw.text(self.title_pos,self.titles[i],**self.title_options)
                    self.image.paste( new_bm.image,
                                      (c * width/cols + tilew/2 - bmw/2 + self.margin,
                                       r * height/rows + tileh/2 - bmh/2 + self.margin) )
                        
                else:
                    break


class DrawBitmap(Bitmap):
    """
    Bitmap with primitives drawn for each unit
    The input matrix has a list of primitives and relative arguments
    for each unit.
    """

    draw_options	= {
    		"fill":		"DarkGray",	# note that is lighter than Gray!
		"width":	1
    }

    def __init__(self, primitive_matrix, box_size ):
        """The overall shape is derived by the sheet shape and the desired
	magnification"""

	border		= 1
        shape		= primitive_matrix.shape
	width		= box_size * shape[ 0 ]
	height		= box_size * shape[ 1 ]
#	seg_len		= int( ( box_size - border ) / 2 ) - 1

        self.image	= Image.new( 'RGB', ( width, height ), 'white' )
        dr_img		= ImageDraw.Draw( self.image )

        for x in range( shape[ 0 ] ):
	    bx		= x * box_size
            for y in range( shape[ 1 ] ):
                by		= y * box_size
		b0		= ( bx + border, by + border )
		b1		= ( bx + box_size - border, by + box_size - border )
                dr_img.rectangle( [ b0, b1 ], fill = self.draw_options[ 'fill' ] )
		for p in primitive_matrix[ y, x ]:
		    p_name	=  p.keys()[ 0 ]
		    if not p_name in dir( dr_img ):
		        raise NotImplementedError( p_name + ' is not a valid draw directive' )
		    val		= p[ p_name ]
		    arg		= self.__in_box( val[ 0 ], b0, box_size - 2 * border )
		    opts	= val[ 1 ]
		    getattr( dr_img, p_name )( arg, **opts )

        

    def __in_box( self, coordinates, box_corner, seg_len ):
        """convert normalized coordinates into image coordinates in the given
	unit box"""

	in_box_coords	= []
	for xy in coordinates:
	    in_box_coords.append( (
	    		box_corner[ 0 ] + seg_len * xy[ 0 ],
			box_corner[ 1 ] + seg_len * xy[ 1 ]
	    ) )
	return in_box_coords


__all__ = [
    "Bitmap",
    "PaletteBitmap",
    "HSVBitmap",
    "RGBBitmap",
    "MontageBitmap",
    "DrawBitmap",
]



"""
General utility functions and classes for Topographica that require numpy.

$Id$
"""

import re

from numpy import sqrt,dot,arctan2,array2string,fmod,floor,array, \
     unravel_index,concatenate,set_printoptions,divide,maximum,minimum
from numpy import ufunc

import param

# Ask numpy to print even relatively large arrays by default
set_printoptions(threshold=200*200)


def ufunc_script_repr(f,imports,prefix=None,settings=None):
    """
    Return a runnable representation of the numpy ufunc f, and an
    import statement for its module.
    """
    # (could probably be generalized if required, because module is
    # f.__class__.__module__)
    imports.append('import numpy')
    return 'numpy.'+f.__name__

from param import parameterized
parameterized.script_repr_reg[ufunc]=ufunc_script_repr


def L2norm(v):
    """
    Return the L2 norm of the vector v.
    """
    return sqrt(dot(v,v))


def divisive_normalization(weights):
    """Divisively normalize an array to sum to 1.0"""
    s = weights.sum()
    if s != 0:
        factor = 1.0/s
        weights *= factor


def add_border(matrix,width=1,value=0.0):
    """   
    Returns a new matrix consisting of the given matrix with a border
    or margin of the given width filled with the given value.
    """
    rows,cols = matrix.shape

    hborder = array([ [value]*(cols+2*width) ]*width)
    vborder = array([ [value]*width ] * rows)

    temp = concatenate( (vborder,matrix,vborder), axis=1)
    return concatenate( (hborder,temp,hborder) )     


def arg(z):
    """
    Return the complex argument (phase) of z.
    (z in radians.)
    """
    z = z + complex(0,0)  # so that arg(z) also works for real z
    return arctan2(z.imag, z.real)


def octave_str(mat,name="mat",owner=""):
    """
    Print the given Numpy matrix in Octave format, listing the given
    matrix name and the object that owns it (if any).
    """
    # This just prints the string version of the matrix and does search/replace
    # to convert it; there may be a faster or easier way.
    mstr=array2string(mat)
    mstr=re.sub('\n','',mstr)
    mstr=re.sub('[[]','',mstr)
    mstr=re.sub('[]]','\n',mstr)
    return ("# Created from %s %s\n# name: %s\n# type: matrix\n# rows: %s\n# columns: %s\n%s" %
           (owner,name,name,mat.shape[0],mat.shape[1],mstr))


def octave_output(filename,mat,name="mat",owner=""):
    """Writes the given matrix to a new file of the given name, in Octave format."""
    f = open(filename,'w')
    f.write(octave_str(mat,name,owner))
    f.close()


def centroid(array_2D):
    """Return the centroid (center of gravity) for a 2D array."""

    rows,cols = array_2D.shape
    rsum=0
    csum=0
    rmass_sum=0
    cmass_sum=0
    for r in xrange(rows):
        row_sum = array_2D[r,:].sum()
        rsum += r*row_sum
        rmass_sum += row_sum
    
    for c in xrange(cols):
        col_sum = array_2D[:,c].sum()
        csum += c*col_sum
        cmass_sum += col_sum
        
    row_centroid= rsum/rmass_sum
    col_centroid= csum/cmass_sum

    return row_centroid, col_centroid


def clip_lower(arr,lower_bound):
    """
    In-place, one-sided version of numpy.clip().

    i.e. numpy.clip(arr,a_min=lower_bound,out=arr) if it existed.
    """
    maximum(arr,lower_bound,arr)


def clip_upper(arr,upper_bound):
    """
    In-place, one-sided version of numpy.clip().

    i.e. numpy.clip(arr,a_max=upper_bound,out=arr) if it existed.
    """
    minimum(arr,upper_bound,arr)


def wrap(lower, upper, x):
    """
    Circularly alias the numeric value x into the range [lower,upper).

    Valid for cyclic quantities like orientations or hues.
    """
    #I have no idea how I came up with this algorithm; it should be simplified.
    #
    # Note that Python's % operator works on floats and arrays;
    # usually one can simply use that instead.  E.g. to wrap array or
    # scalar x into 0,2*pi, just use "x % (2*pi)".
    range_=upper-lower
    return lower + fmod(x-lower + 2*range_*(1-floor(x/(2*range_))), range_)


def array_argmax(arr):
    "Returns the coordinates of the maximum element in the given array."
    return unravel_index(arr.argmax(),arr.shape)



# CB: Is this of general interest? Used in gcal.ty.
class DivideWithConstant(param.Parameterized):
    """
    Divide two scalars or arrays with a constant (c) offset on the
    denominator to allow setting the gain or to avoid divide-by-zero
    issues.  The non-constant part of the denominator (y) is clipped
    to ensure that it has only positive values.
    """
    c = param.Number(default=1.0)
    
    def __call__(self, x, y):
        return divide(x,maximum(y,0)+self.c)



"""
Miscellaneous functions used by Topographica code outside of topo.base.

Several of these files are completely self-contained, not importing
anything from Topographica, and can thus be copied and used in
unrelated projects.

$Id$
"""

# For backwards compatibility; this file used to be in misc/
from imagen import patternfn # pyflakes:ignore (API import)





"""
Unit Support: Classes and methods to allow Unum and Quantities unit
packages to interact with Topographica.

$Id$
"""


import param
from param.parameterized import bothmethod

from topo import sheet, pattern, base, numbergen
import numpy as np

got_unum = False; got_pq=False

try:  
    import quantities as pq 
    got_pq=True
except: pass

try: 
    import unum 
    import unum.units
    unum.Unum.UNIT_FORMAT = '%s'
    got_unum=True
except: pass

if True not in [got_unum, got_pq]: raise ImportError('Could not find Quantities or Unum.')


class strip_pq_hook(param.ParameterizedFunction):
    def __call__(self,obj,val):
        """
        Hook to convert units provided by the quantities package to the base
        unit as defined in the QuantityConversions class and return a float.
        """
        if not hasattr(val,'units'):
            return val
        
        if hasattr(obj,'unit_conversions'):
            return obj.unit_conversions.convert_to_base(val)
        elif hasattr(obj,'src') and hasattr(obj.src,'unit_conversions'):
            return obj.src.unit_conversions.convert_to_base(val)


class strip_unum_hook(param.ParameterizedFunction):
    def __call__(self,obj,val):
        """
        Hook to convert unum unit objects to the specified base unit and
        return as a float.
        """
        if not val.__class__.__name__ == 'Unum':
            return val
        
        if hasattr(obj,'unit_conversions'):
            return obj.unit_conversions.convert_to_base(val)
        elif hasattr(obj,'src') and hasattr(obj.src,'unit_conversions'):
            return obj.src.unit_conversions.convert_to_base(val)



class Conversions(object):
    """
    The UnumConversion class holds unit conversions and definitions,
    setting and resetting the unum unit table allowing sheet specific
    conversions to be performed.
    """

    package = 'Quantities'

    initialized = False

    def __init__(self,units=None):
        """
        Initialize this object first by storing the unit
        specification, backing up the current unit table, creating the
        specified unum unit objects and then restoring the backed up
        unit table.
        """
        if not hasattr(self,'_unit_objects'):
            self._unit_objects = {}
        if not hasattr(self,'_unit_specs'):
            self._unit_specs = {}
        
        if 'unit_conversions' not in sheet.Sheet.params():
            sheet.Sheet._add_parameter("unit_conversions",param.Parameter(None))
        if 'unit_conversions' not in pattern.PatternGenerator.params():
            pattern.PatternGenerator._add_parameter("unit_conversions",param.Parameter(None))            
        if 'unit_conversions' not in numbergen.NumberGenerator.params():
            numbergen.NumberGenerator._add_parameter("unit_conversions",param.Parameter(None))
        
        self.initialize(units)
        self.initialized = True

        
    @bothmethod
    def declare_unit(obj,unit_key,conversion,name,base=False):
        """
        Create and return unit using specified package.
        """
        if obj.package == 'Quantities':
            unit_obj = pq.UnitQuantity(name, definition=conversion,symbol=unit_key)
            
            if base:
                unit = [(unit_obj,unit_key,conversion,name)]
                obj._set_base_units_pq(unit)
            else:
                unit = [(unit_obj,conversion)]
                obj.initialize_units(unit)
            
            return unit_obj

        if obj.package == 'Unum':
            if conversion is None:
                conversion = 0

            obj.del_unit(unit_key)
            unit_obj = unum.Unum.unit(unit_key,conversion,name)
            
            if base:
                unit = [(unit_obj,unit_key,conversion,name)]
                obj._set_base_units_unum(unit)
            else:
                unit = [(unit_obj,conversion)]
                obj.initialize_units(unit)
            
            return unit_obj


    @bothmethod
    def get_unit(obj,unit_key,local=True,global_=True):
        """
        Looks up and returns unit_key in local then global unit
        dictionary, if not in either returns None.
        """
        unit = None
        if local and obj.initialized:
            unit = obj._get_local_unit(unit_key)
        if global_ and not unit:
            unit = obj._get_global_unit(unit_key)
        return unit


    @classmethod
    def set_package(obj,package):
        """
        Set the public methods in accordance with the selected package.
        """
        obj.package = package

        public_methods = ['convert_to_base','initialize','initialize_units','set_local_units']
        unum_methods = [obj._convert_to_base_unum,obj._initialize_unum,obj._initialize_units_unum,obj._set_local_units_unum]
        pq_methods = [obj._convert_to_base_pq,obj._initialize_pq,obj._initialize_units_pq,obj._set_local_units_pq]

        if obj.package == 'Unum' and not got_unum: 
            raise ImportError('Unum package not installed, call Conversions.set_package(\'Quantities\') or install Unum.')
        if obj.package == 'Quantities' and not got_pq: 
            raise ImportError('Quantities package not installed, call Conversions.set_package = (\'Unum\') or install Quantities.')

        if obj.package =='Unum': selected = unum_methods
        else: selected = pq_methods
        [setattr(obj,pub,sel) for (sel,pub) in zip(selected, public_methods)]


    def _convert_to_base_pq(self,val):
        """
        Set local unit definitions and detect unit type then convert
        the value and return as a float.
        """
        self._set_local_units_pq()
        for idx,unit in enumerate(self._base_units):
            try:
                val.rescale(unit[1])
                break
            except: pass
            if idx == len(self._base_units): print 'Cannot convert {0} to base unit.'.format(val)

        return val.rescale(self._base_units[idx][1]).magnitude


    def _convert_to_base_unum(self,val):
        """
        Convert value to base unit using local unit
        definitions and return a float.
        """
        self._ut_bak = unum.Unum.getUnitTable()
        self._set_local_units_unum()
        try:
            val.checkNoUnit()
            val = float(val)
        except unum.ShouldBeUnitlessError:
            for idx,unit in enumerate(self._base_units):
                try:
                    val.asUnit(unit[0])
                    break
                except: pass
                if idx == len(self._base_units): print 'Cannot convert {0} to base unit.'.format(val)
            val = float(val.asUnit(self._base_units[idx][0])/self._base_units[idx][0])

        unum.Unum.reset(self._ut_bak)

        return val


    @bothmethod
    def _del_unit(obj,unit_key):
        """
        Delete specified unit definition from global unum unit table.
        """
        unit_table = unum.Unum.getUnitTable()
        if unit_key in unit_table.keys():
            del unit_table[unit_key]
        unum.Unum.reset(unit_table)


    @bothmethod
    def _get_local_unit(self,unit_key):
        """
        Returns local unit object or if nonexistent None.
        """
        if unit_key in self._unit_objects.keys():
            return self._unit_objects[unit_key]
        else:
            return None


    @bothmethod
    def _get_global_unit(obj,unit_key):
        if obj.package == 'Unum': return getattr(unum.units,unit_key)
        elif obj.package == 'Quantities': return pq.registry.unit_registry[unit_key]
        else: return None  


    def _initialize_pq(self,units):
        """
        Initialize param and Topographica to deal with Quantities unit
        package by installing appropriate set hook on parameters and
        initializing specified units.
        """
        param.Number.set_hook = strip_pq_hook
        base.boundingregion.BoundingRegionParameter.set_hook = strip_pq_hook

        self.initialize_units(units)
        

    def _initialize_unum(self,units):
        """
        Initialize param and Topographica to deal with Unum unit
        package by installing appropriate set hook on parameters and
        initializing specified units.
        """
        param.Number.set_hook = strip_unum_hook
        base.boundingregion.BoundingRegionParameter.set_hook = strip_unum_hook

        self.initialize_units(units)


    @bothmethod
    def _initialize_units_pq(obj,units):
        """
        Initialize specified units using Quantities unit package.
        """
        if not hasattr(obj,'_unit_objects'):
            obj._unit_objects = {}
        if not hasattr(obj,'_unit_specs'):
            obj._unit_specs = {}

        for unit in units:
            unit_key = unit[0].symbol
            obj._unit_objects[unit_key] = unit[0]
            obj._unit_specs[unit_key] = (unit[1],unit[0].name)


    @bothmethod
    def _initialize_units_unum(obj,units):
        """
        Initialize specified units using Unum unit package.
        """
        if not hasattr(obj,'_unit_objects'):
            obj._unit_objects = {}
        if not hasattr(obj,'_unit_specs'):
            obj._unit_specs = {}

        for unit in units:
            unit_key = unit[0].strUnit()
            obj._unit_objects[unit_key] = unit[0]
            obj._unit_specs[unit_key] = (unit[1],unit[0].getUnitTable()[unit_key][2])


    @bothmethod
    def _set_base_units_pq(obj,units):
        """
        Set base unit, which is used to interface with Topographicas
        coordinate system, using Quantities unit package.
        """
        if not hasattr(obj,'_base_units'):
            obj._base_units = []
        for unit in units:
            obj._base_units.append((unit[0],unit[1],unit[2],unit[3]))


    @bothmethod
    def _set_base_units_unum(obj,units):
        """
        Set base units, which are used to interface with Topographicas
        coordinate system, using Unum unit package.
        """
        if not hasattr(obj,'_base_units'):
            obj._base_units = []
        for unit in units:
            obj._base_units.append((unit[0],unit[1],unit[2],unit[3]))


    def _set_local_units_pq(self):
        """
        Set the local unit definitions according to stored unit definitions and using
        the Quantities unit package.
        """
        for unit in self._unit_specs.keys():
            self._unit_objects[unit]._conv_ref = np.array(self._unit_specs[unit][0].magnitude) * self._unit_specs[unit][0].units.simplified
        for idx,unit in enumerate(self._base_units):    
            self._base_units[idx][0]._conv_ref = np.array(unit[2].magnitude) * unit[2].units.simplified
    

    def _set_local_units_unum(self):
        """
        Set the local unit definitions according to stored unit definitions and using
        the Quantities unit package.
        """
        for unit_key in self._unit_specs.keys():
            self.del_unit(unit_key)
            unit_spec = self._unit_specs[unit_key]
            self._unit_objects[unit_key] = unum.Unum.unit(unit_key,unit_spec[0],unit_spec[1])
        for unit in self._base_units:
            self.del_unit(unit[1])
            unum.Unum.unit(unit[1],unit[2],unit[3])

